"""
The submodule `regression` is part of `minterpy-extras`.

It contains functions and classes for multivariate polynomial regression.
"""

__all__ = []

from . import regression_abc  # noqa
from .regression_abc import *

from . import ordinary_regression
from .ordinary_regression import *

__all__ += regression_abc.__all__
__all__ += ordinary_regression.__all__




"""
This is the core module. It should not be touched unless you know exactly what are you doing.
"""

__all__ = []

from . import multi_index  # noqa
from .multi_index import *  # noqa

__all__ += multi_index.__all__

from . import grid  # noqa
from .grid import *  # noqa

__all__ += grid.__all__

from . import ABC  # noqa # ABCs are not exposed to the top level!
from . import tree  # noqa
from . import utils  # noqa # core utils are not exposed to top level



"""
This `transformation` submodule is part of `minterpy`.
"""


__all__ = []

from . import canonical  # noqa
from .canonical import *  # noqa

__all__ += canonical.__all__

from . import newton  # noqa
from .newton import *  # noqa

__all__ += newton.__all__

from . import lagrange  # noqa
from .lagrange import *  # noqa

__all__ += lagrange.__all__

from . import identity  # noqa
from .identity import *  # noqa

__all__ += identity.__all__

from . import interface  # noqa
from .interface import *  # noqa

__all__ += interface.__all__

from . import utils  # noqa # utils are not exposed to toplevel!



"""
Module with functions required for the divided difference scheme in multiple dimensions.

Notes
-----
The barycentric transformations provide a similar functionality (L2N transformation) but with reduced time complexity!
This module also includes the functionality for implicitly creating and traversing a "Multi Index Tree".
The tree structure is being encoded by numpy arrays for increased performance.
"""

from typing import TYPE_CHECKING, Optional, no_type_check

import numpy as np
from numba import njit
from numba.typed import List

from minterpy.global_settings import (
    ARRAY,
    ARRAY_DICT,
    INT_DTYPE,
    INT_SET,
    INT_TUPLE,
    TYPED_LIST,
)

if TYPE_CHECKING:
    from .core.tree import MultiIndexTree


def compile_splits(exponents: ARRAY) -> TYPED_LIST:
    """Identify all the split positions in a multi-index set.

    A split in a given dimension of a multi-index set is indicated
    by a change of value (a "jump") of an exponent one dimension higher.
    The split position is taken after the jump takes place.

    Notes
    -----
    - The first split position is always set to [0].
    - The last dimension contains no split and the split position
      is taken to be [0].
      In other words, one-dimensional multi-index set has no split
      with a split position [0].
    - The splits in a lower dimension includes all the splits
      in the higher dimensions.

    Parameters
    ----------
    exponents : ARRAY
        An array of exponents of a multi-index set.

    Returns
    -------
    TYPED_LIST
        A list of all the split positions for the given multi-index set.
        The length of the list is the number of spatial dimension in
        the set (or the number of columns).

    Examples
    --------
    >>> xx = np.array([
    ...      [0, 0, 0],
    ...      [1, 0, 0],
    ...      [2, 0, 0],
    ...      [0, 1, 0],
    ...      [1, 1, 0],
    ...      [0, 2, 0],
    ...      [0, 0, 1],
    ...      [1, 0, 1],
    ...      [0, 1, 1],
    ...      [1, 1, 1],
    ...      [0, 0, 2],
    ... ])
    >>> compile_splits_numpy(xx)
    ListType[array(int32, 1d, C)]([[ 0  3  5  6  8 10], [ 0  6 10], [0]])
    """

    # Initialize the list of split positions
    split_positions = List()
    # The last element has no split (split position [0])
    split_positions.append(np.array([0], dtype=INT_DTYPE))
    # NOTE: the first split position is always [0]
    prev_split = np.array([0], dtype=INT_DTYPE)

    # A split is indicated when the exponents change value over the rows
    found_splits = exponents[:-1, :] != exponents[1:, :]

    # Iterate over dimension up to the 2nd
    # and find the splits in each dimension
    for i in range(exponents.shape[1] - 1, 0, -1):
        # NOTE: Shift by one to get after the shift takes place not before.
        # np.where returns np.int64,
        # so make sure it's consistent with what Numba expects
        found_split = np.where(found_splits[:, i])[0].astype(INT_DTYPE) + 1
        # Splits in higher dimension is included in the lower dimension
        union = np.union1d(prev_split, found_split)
        prev_split = union
        split_positions.insert(0, union)

    return split_positions


def compile_subtree_sizes(nr_exponents: int, split_positions: TYPED_LIST) -> TYPED_LIST:
    """Compute the sizes of all subtrees in a multi-index tree.

    Each dimension in the multi-index tree contains one or more subtrees (or nodes).
    The size of a subtree refers to the number of the multi-index set elements
    that belongs to that subtree.

    Notes
    -----
    - The largest subtree, that contains all the dimensions, has all the elements
      of the multi-index set. This subtree is the last element of the list.
    - The term subtree and node may be used interchangeably.

    Parameters
    ----------
    nr_exponents : int
        The number of elements in the corresponding multi-index set.
    split_positions: TYPED_LIST
        The list of all the split positions in the corresponding
        multi-index set.

    Returns
    -------
    TYPED_LIST
        A list of all the subtree sizes for the given multi-index tree.
        The length of the list equals to the number of spatial dimension
        in the corresponding multi-index set (or the number of columns).

    Examples
    --------
    >>> xx = np.array([
    ...      [0, 0, 0],
    ...      [1, 0, 0],
    ...      [2, 0, 0],
    ...      [0, 1, 0],
    ...      [1, 1, 0],
    ...      [0, 2, 0],
    ...      [0, 0, 1],
    ...      [1, 0, 1],
    ...      [0, 1, 1],
    ...      [1, 1, 1],
    ...      [0, 0, 2],
    ... ])
    >>> split_positions = compile_splits_numpy(xx)
    >>> compile_subtree_sizes_numpy(len(xx), split_positions)
    ListType[array(int32, 1d, C)]([[3 2 1 2 2 1], [6 4 1], [11]])
    """

    # Initialize the list of subtree sizes
    subtree_sizes = List()

    # Iterate over split positions in each dimension, excluding the last
    for split_position in split_positions[:-1]:
        # NOTE: Include the total number of exponents
        # to get the size of the last subtree in this dimension
        subtree_size = np.diff(split_position, append=nr_exponents)
        subtree_sizes.append(subtree_size)

    # The last element of the list is the largest subtree
    # NOTE: Make sure the array has the correct data type
    subtree_sizes.append(np.array([nr_exponents], dtype=INT_DTYPE))

    return subtree_sizes


def compile_problem_sizes(subtree_sizes: TYPED_LIST) -> TYPED_LIST:
    """Computes the sub-problem sizes for each subtree in the multi-index tree.

    The sub-problem size of a given dimension corresponds to the number
    of subtrees (or nodes) that belong to a single subtree of the higher dimension.

    Notes
    -----
    - The sub-problem size in one-dimension is the same as the subtree size
      in one-dimension.

    Parameters
    ----------
    subtree_sizes: TYPED_LIST
        The list of all the subtree sizes for the given multi-index tree.

    Returns
    -------
    TYPED_LIST
        A list of the problem sizes for the given multi-index tree.
        The length of the list equals to the number of spatial dimension
        in the corresponding multi-index set (or the number of columns).

    Examples
    --------
    >>> xx = np.array([
    ...      [0, 0, 0],
    ...      [1, 0, 0],
    ...      [2, 0, 0],
    ...      [0, 1, 0],
    ...      [1, 1, 0],
    ...      [0, 2, 0],
    ...      [0, 0, 1],
    ...      [1, 0, 1],
    ...      [0, 1, 1],
    ...      [1, 1, 1],
    ...      [0, 0, 2],
    ... ])
    >>> nr_exponents = len(xx)
    >>> split_positions = compile_splits_numpy(xx)
    >>> subtree_sizes = compile_subtree_sizes_numpy(nr_exponents, split_positions)
    >>> compile_problem_sizes_numpy(subtree_sizes)
    ListType[array(int32, 1d, C)]([[3 2 1 2 2 1], [3 2 1], [3]])
    """

    # Initialize the list of problem sizes
    problem_sizes = List()

    # The first problem size is the first subtree size
    problem_sizes.append(subtree_sizes[0])

    subtree_size_child = subtree_sizes[0]
    # Iterate over parent (i.e., higher dimension) subtrees
    for subtree_size_parent in subtree_sizes[1:]:
        problem_size = np.digitize(
            np.cumsum(subtree_size_parent),
            np.cumsum(subtree_size_child)
        ).astype(INT_DTYPE)

        # NOTE: Include 0 at the beginning
        # to get the first problem size in this dimension
        problem_sizes.append(
            np.diff(problem_size, prepend=0)
        )

        subtree_size_child = subtree_size_parent

    return problem_sizes


@njit(cache=True)
def get_node_position(dim_idx: int, node_idx: int, split_positions: TYPED_LIST) -> int:
    """Returns the position of the initial exponent entry corresponding to this node.

    :param dim_idx: dimension
    :param node_idx: node index
    :param split_positions: list of all split positions
    :return: the position of the initial exponent entry corresponding to the node

    Notes
    -----
    Due to JIT compilation this might return unexpected results when the input indices are out of bounds!
    """
    return split_positions[dim_idx][node_idx]


@njit(cache=True)
def get_node_size(dim_idx: int, node_idx: int, subtree_sizes: TYPED_LIST) -> int:
    """Returns the size of the specified node.

    :param dim_idx: dimension
    :param node_idx: node index
    :param subtree_sizes: list of sizes for all subtrees
    :return: problem size of the specified node

    Notes
    -----
    This is equal to the amount of exponents belonging to the corresponding slice/split of the exponent matrix.
    This is equal to the amount of "leaf" nodes which belong to the subtree with the specified node as root node.
    Due to JIT compilation this might return unexpected results when the input indices are out of bounds!
    """
    return subtree_sizes[dim_idx][node_idx]


@njit(cache=True)
def get_positions(
    dim_idx: int, node_idx: int, split_positions: TYPED_LIST, subtree_sizes: TYPED_LIST
) -> INT_TUPLE:
    """Returns the range of indices in the coeffs array that corresponds to the given node index.

    :param dim_idx: dimension
    :param node_idx: node index
    :param split_positions: all split positions
    :param subtree_sizes: all sub tree sizes
    :return: the start and end positions in the coeffs array corresponding to this sub problem
    """
    pos_from = get_node_position(dim_idx, node_idx, split_positions)
    size = get_node_size(dim_idx, node_idx, subtree_sizes)
    pos_to = (
        pos_from + size
    )  # NOTE: this position is NOT included in the referenced "split". used for slicing
    return pos_from, pos_to


@njit(cache=True)
def get_child_idxs(
    dim_idx: int,
    parent_node_idx: int,
    layer_idx: int,
    split_positions: TYPED_LIST,
    subtree_sizes: TYPED_LIST,
) -> INT_TUPLE:
    """Computes the start and end indices of all nodes in the respective layer belonging to the given subtree.

    :param dim_idx: dimension
    :param parent_node_idx: index of the parent node
    :param layer_idx: the index of the layer (= dimension) to search in
    :param split_positions: all split positions
    :param subtree_sizes: all sub tree sizes
    :return: the indices of the first and last node included in the subtree of the "query root node"

    """
    pos_from, pos_to = get_positions(
        dim_idx, parent_node_idx, split_positions, subtree_sizes
    )
    splits2search = split_positions[layer_idx]
    first_child_idx, last_child_idx = (
        0,
        len(splits2search) - 1,
    )  # end at the last entry by default
    search_start = True
    for i, pos in enumerate(splits2search):
        if search_start and pos >= pos_from:
            first_child_idx = i
            search_start = False
        elif pos >= pos_to:  # search_end = not search_start
            last_child_idx = i - 1  # the previous entry is the end of the slice
            break

    return first_child_idx, last_child_idx


@njit(cache=True)
def get_direct_child_idxs(
    dim_idx: int,
    parent_node_idx: int,
    split_positions: TYPED_LIST,
    subtree_sizes: TYPED_LIST,
):
    """Computes the start and end indices of all direct child nodes.

    :param dim_idx: dimension
    :param parent_node_idx: index of the parent node
    :param split_positions: all split positions
    :param subtree_sizes: all sub tree sizes

    """
    child_layer = dim_idx - 1  # next "lower" layer
    return get_child_idxs(
        dim_idx, parent_node_idx, child_layer, split_positions, subtree_sizes
    )


@njit(cache=True)
def get_array_slice(
    dim_idx: int,
    node_idx: int,
    array: ARRAY,
    split_positions: TYPED_LIST,
    subtree_sizes: TYPED_LIST,
) -> ARRAY:
    """Returns the slice of the array corresponding to the given sub problem.

    :param dim_idx: dimension
    :param node_idx: node index
    :param array: the coefficients array
    :param split_positions: all split positions
    :param subtree_sizes: all sub tree sizes
    :return: the slice of the coeffs array that correspond to the given sub problem

    """
    pos_from, pos_to = get_positions(dim_idx, node_idx, split_positions, subtree_sizes)
    return array[pos_from:pos_to]


@njit(cache=True)
def get_node_positions(
    dim_idx: int,
    node_idx: int,
    layer_idx: int,
    split_positions: TYPED_LIST,
    subtree_sizes: TYPED_LIST,
):
    """Returns node positions for all the child nodes.

    :param dim_idx: dimension
    :param node_idx: node index
    :param layer_idx: layer index
    :param split_positions: all split positions
    :param subtree_sizes: all sub tree sizes
    :return: node positions of all the child nodes

    """
    start_idx, end_idx = get_child_idxs(
        dim_idx, node_idx, layer_idx, split_positions, subtree_sizes
    )
    # NOTE: in order to include the last node as expected while slicing, increment the last index by 1
    end_idx += 1
    splits_in_layer = split_positions[layer_idx]
    selected_splits = splits_in_layer[start_idx:end_idx]
    return selected_splits


@njit(cache=True)
def get_leaf_idxs(
    dim_idx: int, node_idx: int, split_positions: TYPED_LIST, subtree_sizes: TYPED_LIST
):
    """Returns node indices of the leaf nodes in the given sub tree.

    :param dim_idx: dimension
    :param node_idx: node index
    :param split_positions: all split positions
    :param subtree_sizes: all sub tree sizes
    :return: the node indices of leaf nodes in the given sub tree

    """
    leaf_layer = 0  # the "lowest" layer
    start_idx, end_idx = get_child_idxs(
        dim_idx, node_idx, leaf_layer, split_positions, subtree_sizes
    )
    # NOTE: in order to include the last node as expected while slicing, increment the last index by 1
    end_idx += 1
    return start_idx, end_idx


@njit(cache=True)
def get_nodes(
    dim_idx: int,
    node_idx: int,
    layer_idx: int,
    split_positions: TYPED_LIST,
    subtree_sizes: TYPED_LIST,
):
    """Returns the nodes and corresponding problem sizes.

    :param dim_idx: dimesion
    :param node_idx: node index
    :param layer_idx: layer index
    :param split_positions: all split positions
    :param subtree_sizes: all sub tree sizes
    :return: the node indices and their respective problem sizes

    """
    start_idx, end_idx = get_child_idxs(
        dim_idx, node_idx, layer_idx, split_positions, subtree_sizes
    )
    # NOTE: in order to include the last node as expected while slicing, increment the last index by 1
    end_idx += 1
    splits_in_layer = split_positions[layer_idx]
    sizes_in_layer = subtree_sizes[layer_idx]
    selected_splits = splits_in_layer[start_idx:end_idx]
    selected_sizes = sizes_in_layer[start_idx:end_idx]
    return selected_splits, selected_sizes


@njit(cache=True)
def get_leaves(
    dim_idx: int, node_idx: int, split_positions: TYPED_LIST, subtree_sizes: TYPED_LIST
):
    """Returns all the nodes and corresponding problem sizes.

    :param dim_idx: dimesion
    :param node_idx: node index
    :param split_positions: all split positions
    :param subtree_sizes: all sub tree sizes
    :return: all node indices and their respective problem sizes

    """
    leaf_layer = 0  # the "lowest" layer
    selected_splits, selected_sizes = get_nodes(
        dim_idx, node_idx, leaf_layer, split_positions, subtree_sizes
    )
    return selected_splits, selected_sizes


@njit(cache=True)
def compute_projection_mask(
    dim_idx: int,
    node_idx_left: int,
    node_idx_right: int,
    split_positions: TYPED_LIST,
    subtree_sizes: TYPED_LIST,
    exponents: ARRAY,
) -> Optional[ARRAY]:
    """Find the correspondences between exponent entries in the left and the right node.

    :param dim_idx: dimension
    :param node_idx_left: the index of the left node to be considered
    :param node_idx_right: the index of the right node to be considered
    :param split_positions: all split positions
    :param subtree_sizes: all sub tree sizes
    :param exponents: the exponents array
    :return: the indices of the entries in the left subtree split
             corresponding to all entries in the right subtree split
    """

    # NOTE: the left subtree is often much larger than the right subtrees
    #     -> avoid creating a very sparse boolean mask
    # instead find the indices of the left leaves (segmented) corresponding to all right leaves (continuous)
    # (store only the matching idxs)
    #
    # NOTE: a "left" subtree is always bigger than any subtrees to its right
    #
    # this implementation exploits another special property of complete index sets:
    #     the corresponding leaves in the left tree will be completely contained in the right subtree
    #     and due to the lexicographical ordering the first entries of the leaves will match!
    #
    # NOTE: there might be more left leaves as right leaves! -> not a 1:1 correspondence between the leaf nodes!
    #     -> correspondence check via the exponents required!

    tree_size_r = get_node_size(dim_idx, node_idx_right, subtree_sizes)
    if (
        tree_size_r == 1
    ):  # only a single leaf entry (= all 0) also always corresponds to the first entry!
        mask = np.zeros(1, dtype=INT_DTYPE)
        return mask

    tree_size_l = get_node_size(dim_idx, node_idx_left, subtree_sizes)
    # when the two problems have the same size, the correspondence is trivial (1:1) -> no modification needed
    if tree_size_l == tree_size_r:
        mask = np.zeros(0, dtype=INT_DTYPE)  # default
        return mask

    mask = np.empty(tree_size_r, dtype=INT_DTYPE)

    leaf_positions_l, leaf_sizes_l = get_leaves(
        dim_idx, node_idx_left, split_positions, subtree_sizes
    )
    leaf_positions_r, leaf_sizes_r = get_leaves(
        dim_idx, node_idx_right, split_positions, subtree_sizes
    )
    nr_of_leaves_r = len(leaf_sizes_r)
    nr_entries_matched_r = 0
    nr_entries_matched_l = 0
    leaf_nr_r = 0
    dim_idx = (
        dim_idx + 1
    )  # required for slicing the exponent vectors for comparison correctly
    leaf_pos_r = leaf_positions_r[leaf_nr_r]  # index of the first entry of this leaf
    exp_vect_r = exponents[leaf_pos_r, :dim_idx]
    for leaf_pos_l, leaf_size_l in zip(leaf_positions_l, leaf_sizes_l):
        # NOTE: some left leaf nodes do not have a correspondence in the right subtree!
        # -> check for equality of the first entry (only in the lower dimensions)
        exp_vect_l = exponents[leaf_pos_l, :dim_idx]
        if np.array_equal(exp_vect_r, exp_vect_l):  # -> these two leaf nodes match
            # link all positions in the left leaf corresponding to existing positions in the right leaf:
            leaf_size_r = leaf_sizes_r[leaf_nr_r]
            for i in range(leaf_size_r):
                mask[nr_entries_matched_r] = nr_entries_matched_l + i
                nr_entries_matched_r += 1
            leaf_nr_r += 1
            if leaf_nr_r == nr_of_leaves_r:  # found all relevant correspondences. abort
                break
            leaf_pos_r = leaf_positions_r[leaf_nr_r]
            exp_vect_r = exponents[leaf_pos_r, :dim_idx]

        nr_entries_matched_l += leaf_size_l

    return mask


@njit(cache=True)
def precompute_masks(
    split_positions: TYPED_LIST, subtree_sizes: TYPED_LIST, exponents: ARRAY
) -> ARRAY_DICT:
    """Computes and stores all required correspondences between nodes in the left and the right of the tree
    based on the given splitting.

    :param split_positions: all split positions
    :param subtree_sizes: all sub tree sizes
    :param exponents: exponents array
    :return: a dictionary of all mappings from left to right
    """
    masks = dict()  # required Numba numba.typed.Dict
    dimensionality = len(split_positions)
    # recurse through the tree
    # NOTE: no masks are required for the last dimension (1D DDS used)
    for dim_idx_par in range(
        dimensionality - 1, 0, -1
    ):  # starting from the highest dimension
        dim_idx_child = dim_idx_par - 1
        splits_in_dim = split_positions[dim_idx_par]
        nr_nodes_in_dim = len(splits_in_dim)
        for node_idx_par in range(nr_nodes_in_dim):  # all parent nodes
            first_child_idx, last_child_idx = get_direct_child_idxs(
                dim_idx_par, node_idx_par, split_positions, subtree_sizes
            )
            if first_child_idx == last_child_idx:
                # ATTENTION: left and right node must not be equal!
                continue
            for node_idx_l in range(first_child_idx, last_child_idx):
                for node_idx_r in range(
                    node_idx_l + 1, last_child_idx + 1
                ):  # ATTENTION: also include the last idx!
                    mask = compute_projection_mask(
                        dim_idx_child,
                        node_idx_l,
                        node_idx_r,
                        split_positions,
                        subtree_sizes,
                        exponents,
                    )
                    masks[(dim_idx_child, node_idx_l, node_idx_r)] = mask

    return masks


@njit(cache=True)
def dds_1_dimensional(grid_values: ARRAY, result_placeholder: ARRAY) -> None:
    """One dimensional divided difference scheme

    :param grid_values: the positions of the interpolation nodes respectively
    :param result_placeholder: 2D array initially containing the function values
    :return: the Newton coefficients corresponding to the interpolating 1D polynomial(s)

    Notes
    -----
    Works on a result placeholder. Nothing is returned.
    Also works with multiple sets of given input coefficients.

    """

    # based on:
    #     https://en.wikipedia.org/wiki/Divided_differences
    #     https://stackoverflow.com/questions/14823891/newton-s-interpolating-polynomial-python
    #
    # TODO often used with triangular input -> triangular output
    #     optimise for this use case

    c = result_placeholder  # alias
    n = len(c)
    # NOTE: the function values get split up. the grid values however do not!
    v = grid_values[:n]  # use only the first n grid values
    for i in range(1, n):
        # IMPORTANT: assign/replace values of array view (slice)!
        i_prev = i - 1
        coeff_slice = c[i:]
        val_slice = v[i:]
        val_diff = val_slice - v[i_prev]
        val_diff = np.expand_dims(val_diff, -1)
        coeff_diff = coeff_slice - c[i_prev]
        coeff_slice[:] = coeff_diff / val_diff


@no_type_check
@njit(cache=True)
def project_n_update(
    dim_idx: int,
    node_idx_l: int,
    node_idx_r: int,
    exponent_l: int,
    v_left: ARRAY,
    result_placeholder: ARRAY,
    generating_values: ARRAY,
    split_positions: TYPED_LIST,
    subtree_sizes: TYPED_LIST,
    masks: ARRAY_DICT,
) -> None:
    """Projects v_left onto v_right and computes the divided difference.

    Each split corresponds to a slice of the function values (v_left and v_right).
    Project a value vector slice belonging to a left subtree onto its correspondence of a right subtree.

    :param dim_idx: dimension
    :param node_idx_l: the index of the left node
    :param node_idx_r: the index of the right node
    :param exponent_l: exponents in the left node
    :param v_left: slice of the coeffs array on the left
    :param result_placeholder: placeholder for divided differences
    :param generating_values: generating values used to create interpolation nodes
    :param split_positions: all split positions
    :param subtree_sizes: all subtree sizes
    :param masks: a dictionary of all precomputed required correspondences between left and right nodes

    """
    idx_offset = node_idx_r - node_idx_l
    exponent_r = exponent_l + idx_offset
    # NOTE: iterate over all splits to the right (<-> "right" nodes)

    gen_val_l = generating_values[exponent_l]
    gen_val_r = generating_values[exponent_r]
    grid_val_diff = gen_val_r - gen_val_l

    mask = masks[
        (dim_idx, node_idx_l, node_idx_r)
    ]  # look up the mapping between left and right
    if (
        len(mask) > 0
    ):  # only if the mask contains entries (no mapping required otherwise)
        v_left = v_left[mask]  # project v_left onto each v_right

    # function value corresponding to the right subtree
    v_right = get_array_slice(
        dim_idx, node_idx_r, result_placeholder, split_positions, subtree_sizes
    )
    # L_2 = (L - Q_1) / Q_H
    v_right[:] = (v_right - v_left) / grid_val_diff  # replaces all values of the view


@njit(cache=True)
def jit_dds(
    result_placeholder: ARRAY,
    generating_points: ARRAY,
    split_positions: TYPED_LIST,
    subtree_sizes: TYPED_LIST,
    masks: ARRAY_DICT,
    exponents: ARRAY,
) -> None:
    """Divided difference scheme for multiple dimensions

    :param result_placeholder: 2D array where the results (=Newton coefficients) should be stored.
        Initially this array must contain the function values on the corresponding unisolvent nodes
        (= Lagrange coefficients).
        This array may contain multiple sets of coefficients (=columns of the input matrix!) at once.
    :param generating_points: generating points for the grid
    :param split_positions: all the split positions
    :param subtree_sizes: all sub tree sizes
    :param masks: a dictionary of all precomputed required correspondences between left and right nodes
    :param exponents: the exponents array

    Notes
    -----
    Iterative scheme. Works on a result placeholder. Nothing is returned.
    Finally the ```result_placeholder``` contains the Newton coefficients of the polynomial
    which has the input function values at the respective unisolvent nodes.
    which has the input function values at the respective unisolvent nodes.
    """

    # ATTENTION: the ordering of the recursive (sub-) function calls is relevant!
    #     this is because the algorithms work on the same result placeholder (<- the views of a single array)!

    dimensionality = len(split_positions)
    # traverse through the "tree" (mimicking recursion)
    # NOTE: in the last dimension the regular (1D DDS can be used)
    for dim_idx_par in range(
        dimensionality - 1, 0, -1
    ):  # starting from the highest dimension
        dim_idx_child = dim_idx_par - 1
        splits_in_dim = split_positions[dim_idx_par]
        nr_nodes_in_dim = len(splits_in_dim)
        generating_values = generating_points[:, dim_idx_par]
        for node_idx_par in range(nr_nodes_in_dim):  # for all parent nodes
            first_child_idx, last_child_idx = get_direct_child_idxs(
                dim_idx_par, node_idx_par, split_positions, subtree_sizes
            )
            if first_child_idx == last_child_idx:
                # ATTENTION: left and right node must not be equal!
                continue  # abort when there are no "neighbouring nodes"
            # each split corresponds to a slice of the function values (v_left and v_right)
            for node_idx_l in range(first_child_idx, last_child_idx):
                # function value corresponding to the left subtree
                v_left = get_array_slice(
                    dim_idx_child,
                    node_idx_l,
                    result_placeholder,
                    split_positions,
                    subtree_sizes,
                )
                pos_l = get_node_position(dim_idx_child, node_idx_l, split_positions)
                # look up the "exponent of this sub problem"
                exponent_l = exponents[
                    pos_l, dim_idx_par
                ]  # ATTENTION: from the "dimension above"
                # iterate over all splits to the right (<-> "right" nodes)
                for node_idx_r in range(
                    node_idx_l + 1, last_child_idx + 1
                ):  # ATTENTION: also include the last idx!
                    project_n_update(
                        dim_idx_child,
                        node_idx_l,
                        node_idx_r,
                        exponent_l,
                        v_left,
                        result_placeholder,
                        generating_values,
                        split_positions,
                        subtree_sizes,
                        masks,
                    )

    # compute the usual 1D DDS for ALL leaf nodes!
    dim_idx_child = 0
    splits_in_dim = split_positions[dim_idx_child]
    nr_nodes_in_dim = len(splits_in_dim)
    generating_values = generating_points[:, dim_idx_child]
    for node_idx_par in range(nr_nodes_in_dim):
        v_leaf = get_array_slice(
            dim_idx_child,
            node_idx_par,
            result_placeholder,
            split_positions,
            subtree_sizes,
        )
        dds_1_dimensional(generating_values, v_leaf)


def dds(fct_values: ARRAY, tree: "MultiIndexTree") -> ARRAY:
    """Computes the newton coefficients for the multi dimensional polynomial using divided differences.

    :param fct_values: the function values on the unisolvent nodes
    :param tree: the MultiIndex tree instance
    :return: the newton coefficients of the polynomial

    """
    # TODO type checking?!
    # check_type_n_values(fct_values)
    # check_shape(fct_values, shape=[len(tree.multi_index)])

    # NOTE: for more memory efficiency computes the results "in place"
    # initialise the placeholder with the function values (= Lagrange coefficients
    # NOTE: the DDS function expects a 2D array as input
    result_placeholder = fct_values.copy()
    if fct_values.ndim == 1:
        # ATTENTION: the DDS operates on the first dimension! -> second dimension must be 1
        result_placeholder = result_placeholder.reshape(-1, 1)
    generating_points = tree.grid.generating_points
    split_positions = tree.split_positions
    subtree_sizes = tree.subtree_sizes
    masks = tree.stored_masks
    exponents = tree.multi_index.exponents
    jit_dds(
        result_placeholder,
        generating_points,
        split_positions,
        subtree_sizes,
        masks,
        exponents,
    )
    return result_placeholder  # = Newton coefficients



"""
Matrix operator class.
"""
from minterpy.core.ABC import OperatorABC
from minterpy.global_settings import ARRAY

__all__ = ["MatrixOperator"]


class MatrixOperator(OperatorABC):
    """Concrete implementation of a Operator constructed as a matrix."""

    def __matmul__(self, other):
        if isinstance(other, OperatorABC):
            # the input is another transformation
            # instead of an array return another Matrix transformation operator constructed from the matrix product
            # TODO which transformation object should be passed?
            return MatrixOperator(
                self.transformation, self.array_repr_full @ other.array_repr_full
            )

        return self.array_repr_sparse @ other

    def _get_array_repr(self) -> ARRAY:
        """Returns the matrix representation of the transformation."""
        return self.transformation_data



"""
Abstract base class for polynomial regression classes.

This module contains the abstract base class from which all concrete
implementations of polynomial regression classes should be created.
Such an abstract class ensures all the polynomial regression classes shares a set of common interface,
while their implementation may differs or additional methods or attributes may be defined.
"""

import abc


__all__ = ["RegressionABC"]


class RegressionABC(abc.ABC):
    """The abstract base class for all regression models."""

    @abc.abstractmethod
    def fit(self, xx, yy, *args, **kwargs):  # pragma: no cover
        """Abstract container for fitting a polynomial regression."""
        pass

    @abc.abstractmethod
    def predict(self, xx):  # pragma: no cover
        """Abstract container for making prediction using a polynomial regression."""
        pass

    @abc.abstractmethod
    def show(self):  # pragma: no cover
        """Abstract container for printing out the details of a polynomial regression model."""
        pass

    def __call__(self, xx):  # pragma: no cover
        """Evaluation of the polynomial regression model."""
        return self.predict(xx)



"""
This is the conftest module of minterpy.

Within a pytest run, this module is loaded first. That means here all global fixutes shall be defined.
"""

import inspect

import numpy as np
import pytest
from numpy.testing import assert_, assert_almost_equal, assert_equal

from minterpy import MultiIndexSet, NewtonPolynomial

# Global seed
SEED = 12345678


# asserts that a call runs as expected
def assert_call(fct, *args, **kwargs):
    try:
        fct(*args, **kwargs)
    except Exception as e:
        print(type(e))
        raise AssertionError(
            f"The function was not called properly. It raised the exception:\n\n {e.__class__.__name__}: {e}"
        )


# assert if multi_indices are equal
def assert_multi_index_equal(mi1, mi2):
    try:
        assert_(isinstance(mi1, type(mi2)))
        assert_equal(mi1.exponents, mi2.exponents)
        assert_equal(mi1.lp_degree, mi2.lp_degree)
        assert_equal(mi1.poly_degree, mi2.poly_degree)
    except AssertionError as a:
        raise AssertionError(
            f"The two instances of MultiIndexSet are not equal:\n\n {a}"
        )


# assert if multi_indices are almost equal
def assert_multi_index_almost_equal(mi1, mi2):
    try:
        assert_(isinstance(mi1, type(mi2)))
        assert_almost_equal(mi1.exponents, mi2.exponents)
        assert_almost_equal(mi1.lp_degree, mi2.lp_degree)
        assert_almost_equal(mi1.poly_degree, mi2.poly_degree)
    except AssertionError as a:
        raise AssertionError(
            f"The two instances of MultiIndexSet are not almost equal:\n\n {a}"
        )


# assert if two grids are equal
def assert_grid_equal(grid1, grid2):
    try:
        assert_(isinstance(grid1, type(grid2)))
        assert_equal(grid1.unisolvent_nodes, grid2.unisolvent_nodes)
        assert_equal(grid1.spatial_dimension, grid2.spatial_dimension)
        assert_equal(grid1.generating_values, grid2.generating_values)
        assert_equal(grid1.generating_values, grid2.generating_values)
        assert_multi_index_equal(grid1.multi_index, grid2.multi_index)
    except AssertionError as a:
        raise AssertionError(f"The two instances of Grid are not equal:\n\n {a}")


# assert if two grids are almost equal
def assert_grid_almost_equal(grid1, grid2):
    try:
        assert_(isinstance(grid1, type(grid2)))
        assert_almost_equal(grid1.unisolvent_nodes, grid2.unisolvent_nodes)
        assert_almost_equal(grid1.spatial_dimension, grid2.spatial_dimension)
        assert_almost_equal(grid1.generating_values, grid2.generating_values)
        assert_almost_equal(grid1.generating_values, grid2.generating_values)
        assert_multi_index_almost_equal(grid1.multi_index, grid2.multi_index)
    except AssertionError as a:
        raise AssertionError(f"The two instances of Grid are not almost equal:\n\n {a}")


# assert if polynomials are almost equal
def assert_polynomial_equal(P1, P2):
    try:
        assert_(isinstance(P1, type(P2)))
        assert_multi_index_equal(P1.multi_index, P2.multi_index)
        assert_equal(P1.coeffs, P2.coeffs)
    except AssertionError as a:
        raise AssertionError(
            f"The two instances of {P1.__class__.__name__} are not equal:\n\n {a}"
        )


# assert if polynomials are almost equal
def assert_polynomial_almost_equal(P1, P2):
    try:
        assert_(isinstance(P1, type(P2)))
        assert_multi_index_almost_equal(P1.multi_index, P2.multi_index)
        assert_almost_equal(P1.coeffs, P2.coeffs)
    except AssertionError as a:
        raise AssertionError(
            f"The two instances of {P1.__class__.__name__} are not almost equal:\n\n {a}"
        )


# assert if two functions have the same object code
def assert_function_object_code_equal(fct1, fct2):
    try:
        assert_(fct1.__code__.co_code == fct2.__code__.co_code)
    except AssertionError as a:
        raise AssertionError(
            f"The object_code of {fct1} is not equal to the object_code of {fct2}:\n\n {inspect.getsource(fct1)}\n\n{inspect.getsource(fct2)}"
        )


# assert if interpolators are equal
def assert_interpolator_equal(interpolator1, interpolator2):
    try:
        assert_(isinstance(interpolator1, type(interpolator2)))
        assert_equal(interpolator1.spatial_dimension, interpolator2.spatial_dimension)
        assert_equal(interpolator1.poly_degree, interpolator2.poly_degree)
        assert_equal(interpolator1.lp_degree, interpolator2.lp_degree)
        assert_multi_index_equal(interpolator1.multi_index, interpolator2.multi_index)
        assert_grid_equal(interpolator1.grid, interpolator2.grid)
    except AssertionError as a:
        raise AssertionError(
            f"The two instances of {interpolator1.__class__.__name__} are not equal:\n\n {a}"
        )


# assert if interpolators are almost equal
def assert_interpolator_almost_equal(interpolator1, interpolator2):
    try:
        assert_(isinstance(interpolator1, type(interpolator2)))
        assert_almost_equal(
            interpolator1.spatial_dimension, interpolator2.spatial_dimension
        )
        assert_almost_equal(interpolator1.poly_degree, interpolator2.poly_degree)
        assert_almost_equal(interpolator1.lp_degree, interpolator2.lp_degree)
        assert_multi_index_almost_equal(
            interpolator1.multi_index, interpolator2.multi_index
        )
        assert_grid_almost_equal(interpolator1.grid, interpolator2.grid)
    except AssertionError as a:
        raise AssertionError(
            f"The two instances of {interpolator1.__class__.__name__} are not almost equal:\n\n {a}"
        )


# assert if interpolants are equal
def assert_interpolant_equal(interpolant1, interpolant2):
    try:
        assert_(isinstance(interpolator1, type(interpolator2)))
        assert_function_object_code_equal(interpolant1.fct, interpolant2.fct)
        assert_interpolator_equal(interpolant1.interpolator, interpolant2.interpolator)
    except AssertionError as a:
        raise AssertionError(
            f"The two instances of {interpolant1.__class__.__name__} are not equal:\n\n {a}"
        )


# assert if interpolants are almost equal
def assert_interpolant_almost_equal(interpolant1, interpolant2):
    try:
        assert_(isinstance(interpolant1, type(interpolant2)))
        assert_function_object_code_equal(interpolant1.fct, interpolant2.fct)
        assert_interpolator_almost_equal(
            interpolant1.interpolator, interpolant2.interpolator
        )
    except AssertionError as a:
        raise AssertionError(
            f"The two instances of {interpolant1.__class__.__name__} are not almost equal:\n\n {a}"
        )


# fixtures for spatial dimension

spatial_dimensions = [1, 3]


@pytest.fixture(params=spatial_dimensions)
def SpatialDimension(request):
    return request.param


# fixture for polynomial degree
# NOTE: Include test for poly_degree 0 (Issue #27)
polynomial_degree = [0, 1, 4]


@pytest.fixture(params=polynomial_degree)
def PolyDegree(request):
    return request.param


# fixture for lp degree

lp_degree = [0.5, 1, 2, np.inf]


@pytest.fixture(params=lp_degree)
def LpDegree(request):
    return request.param


# fixtures for multi_indices


@pytest.fixture()
def MultiIndices(SpatialDimension, PolyDegree, LpDegree):
    return MultiIndexSet.from_degree(SpatialDimension, PolyDegree, LpDegree)


# fixtures for number of similar polynomials

nr_similar_polynomials = [None, 1, 2]


@pytest.fixture(params=nr_similar_polynomials)
def NrSimilarPolynomials(request):
    return request.param


# fixture for the number of points for evaluations

nr_pts = [1, 2]


@pytest.fixture(params=nr_pts)
def NrPoints(request):
    return request.param


# Fixture for the number
nr_polynomials = [1, 10]


@pytest.fixture(params=nr_polynomials)
def NrPolynomials(request):
    return request.param


# Fixture for the number
batch_sizes = [1, 100, 1000]


@pytest.fixture(params=batch_sizes)
def BatchSizes(request):
    return request.param

# some random builder


def build_rnd_exponents(dim, n, seed=None):
    """Build random exponents.

    For later use, if ``MultiIndexSet`` will accept arbitrary exponents again.

    :param dim: spatial dimension
    :param n: number of random monomials

    Notes
    -----
    Exponents are generated within the intervall ``[MIN_POLY_DEG,MAX_POLY_DEG]``

    """
    if seed is None:
        seed = SEED
    np.random.seed(seed)
    return np.random.randint(MIN_POLY_DEG, MAX_POLY_DEG, (n, dim))


def build_rnd_coeffs(mi, nr_poly=None, seed=None):
    """Build random coefficients.

    For later use.

    :param mi: The :class:`MultiIndexSet` instance of the respective polynomial.
    :type mi: MultiIndexSet

    :param nr_poly: Number of similar polynomials. Default is 1
    :type nr_poly: int, optional

    :return: Random array of shape ``(nr of monomials,nr of similar polys)`` usable as coefficients.
    :rtype: np.ndarray

    """
    if nr_poly is None:
        additional_coeff_shape = tuple()
    else:
        additional_coeff_shape = (nr_poly,)
    if seed is None:
        seed = SEED
    np.random.seed(seed)
    return np.random.random((len(mi),) + additional_coeff_shape)


def build_rnd_points(nr_points, spatial_dimension, nr_poly=None, seed=None):
    """Build random points in space.

    Return a batch of `nr_points` vectors in the real vector space of dimension ``spatial_dimension``.

    :param nr_points: Number of points
    :type nr_points: int

    :param spatial_dimension: Dimension of domain space.
    :type spatial_dimension: int

    :param nr_poly: Number of similar polynomials. Default is 1
    :type nr_poly: int, optional

    :param seed: Seed used for the random generation. Default is SEED.
    :type seed: int

    :return: Array of shape ``(nr_points,spatial_dimension[,nr_poly])`` containig random real values (distributed in the intervall :math:`[-1,1]`).
    :rtype: np.ndarray

    """
    if nr_poly is None:
        additional_coeff_shape = tuple()
    else:
        additional_coeff_shape = (nr_poly,)
    if seed is None:
        seed = SEED
    np.random.seed(seed)
    return np.random.uniform(
        -1, 1, size=(nr_points, spatial_dimension) + additional_coeff_shape
    )


def build_random_newton_polynom(
    dim: int, deg: int, lp: int, seed=None
) -> NewtonPolynomial:
    """Build a random Newton polynomial.

    Return a :class:`NewtonPolynomial` with a lexicographically complete :class:`MultiIndex` (initiated from ``(dim,deg,lp)``) and randomly generated coefficients (uniformly distributes in the intervall :math:`[-1,1]`).

    :param dim: dimension of the domain space.
    :type dim: int
    :param deg: degree of the interpolation polynomials
    :type deg: int
    :param lp: degree of the :math:`l_p` norm used to determine the `poly_degree`.
    :type lp: int

    :return: Newton polynomial with random coefficients.
    :rtype: NewtonPolynomial

    """
    mi = MultiIndexSet.from_degree(dim, deg, lp)
    if seed is None:
        seed = SEED

    rnd_coeffs = np.random.uniform(-1, 1, size=len(mi))
    return NewtonPolynomial(mi, rnd_coeffs)



"""
Abstract base class for defining transformation operators from one polynomial basis to another.

This module provides the abstract base class for polynomial basis transformation operators from which all concrete
implentations are derived.

.. todo::
    "inverse" property useful?
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from minterpy.global_settings import ARRAY

if TYPE_CHECKING:
    # https://stackoverflow.com/questions/39740632/python-type-hinting-without-cyclic-imports
    from .transformation_abstract import TransformationABC


__all__ = ["OperatorABC"]


class OperatorABC(ABC):
    """Abstract base class for transformation operators.

    All transformation operators must be derived from this base class.

    Attributes
    ----------
    array_repr_full
    array_repr_sparse

    """

    # ATTENTION: this approach is only valid for single polynomials
    transformation_data = None
    _array_repr_full: ARRAY | None = None

    # TODO useful to store the attached bases (Polynomial classes)...
    def __init__(self, transformation: TransformationABC, transformation_data):
        self.transformation: TransformationABC = transformation
        self.transformation_data = transformation_data

    @abstractmethod
    def __matmul__(self, other: ARRAY | OperatorABC):
        """Applies the transformation operator on the input.

        This is a placeholder of the ABC, which is overwritten by the concrete implementation.
        """
        # NOTE: input may be another transformation for chaining multiple transformation operators together
        # TODO input type checking here?
        # TODO discuss: own function for this?
        pass

    @abstractmethod
    def _get_array_repr(self) -> ARRAY:
        """Function computing the full array performing the transformation ("transformation matrix").

        Notes
        -----
        The output should transform the whole basis (not only some active monomials)!
        """
        pass

    @property
    def array_repr_full(self) -> ARRAY:
        """Array representation of the global transformation matrix.

        :return: the matrix representation of the transformation.

        Notes
        -----
        The output transforms the whole basis (not only the active monomials)!
        """
        if self._array_repr_full is None:  # initialise
            self._array_repr_full = self._get_array_repr()
        return self._array_repr_full

    @property
    def array_repr_sparse(self) -> ARRAY:
        """Array representation of the sub-transformation matrix transforming only the active monomials.

        :return: the transformation matrix for the active monomials.

        Notes
        -----
        This is an experimental feature which is part of ongoing work.

        """
        array_repr = self.array_repr_full
        origin_poly = self.transformation.origin_poly
        if origin_poly.indices_are_separate:
            # select only the required columns of the transformation matrix
            # TODO: find a way to make the computation more efficient, avoid computing the "whole" transformation
            active_idxs = origin_poly.active_monomials
            array_repr = array_repr[:, active_idxs]
        return array_repr



"""
Module providing high-level helper functions for polynomial basis transformations.
"""
from __future__ import annotations

from typing import no_type_check

from minterpy.core.ABC import MultivariatePolynomialSingleABC, TransformationABC

from .identity import Identity

__all__ = ["get_transformation", "get_transformation_class"]


@no_type_check
def get_transformation_class(
    origin_type: MultivariatePolynomialSingleABC,
    target_type: type[MultivariatePolynomialSingleABC],
) -> type[TransformationABC]:
    """Finds the Transformation class to go from origin_type to target_type

    A registry of 'available_transforms' is maintained in the TransformationABC. This function performs a lookup in
    this registry to retrieve the correct Transformation class.

    :param origin_type: data type of the origin polynomial
    :param target_type: data type of the target polynomial
    :return: the Transformation class that can perform this transformation.

    """
    if origin_type == target_type:
        return Identity

    if (origin_type, target_type) in TransformationABC.available_transforms.keys():
        return TransformationABC.available_transforms[(origin_type, target_type)]
    else:
        raise NotImplementedError(
            f"There is no known transformation from {origin_type} to {target_type}. \
            \nThe available transformations are {[(key1.__name__, key2.__name__) for (key1, key2) in TransformationABC.available_transforms.keys()]}"
        )


@no_type_check
def get_transformation(
    origin_polynomial: type[MultivariatePolynomialSingleABC],
    target_type: type[MultivariatePolynomialSingleABC],
) -> type[TransformationABC]:
    """Finds the Transformation class that can transform the basis of the origin_polynomial to the desired target_type.

    :param origin_polynomial: an instance of the origin polynomial
    :param target_type: data type of the target polynomial
    :return: the Transformation class that can perform this transformation.

    """
    transformation_class = get_transformation_class(
        origin_polynomial.__class__, target_type
    )
    return transformation_class(origin_polynomial)



"""
Abstract base class for defining transformations from one polynomial basis to another.

This module provides the abstract base class for polynomial basis transformations (origin to target) from
which all concrete implentations are derived.
"""

from abc import ABC, abstractmethod
from typing import Any, Dict, Optional

import numpy as np

from ..grid import Grid
from ..multi_index import MultiIndexSet
from .multivariate_polynomial_abstract import MultivariatePolynomialSingleABC
from .operator_abstract import OperatorABC

__all__ = ["TransformationABC"]


class TransformationABC(ABC):
    """this is the abstract base class for polynomial basis transformations.

    All polynomial basis transformers should be derived from this class to maintain a uniform interface for
     basis transformations.

    Attributes
    ----------
    origin_type
    target_type
    origin_poly : MultivariatePolynomialSingleABC
        The instance of the polynomial to be transformed.
    multi_index
    grid


    """

    available_transforms: Dict[Any, Any] = {}

    def __init__(self, origin_poly: MultivariatePolynomialSingleABC):
        if not isinstance(origin_poly, MultivariatePolynomialSingleABC):
            raise TypeError(f"<{origin_poly}> is not a Polynomial type.")
        if not isinstance(origin_poly, self.origin_type):
            raise TypeError(
                f"<{origin_poly}> is not of the expected type {self.origin_type}."
            )
        self.origin_poly = origin_poly

        # TODO check for index completeness
        # is complete: -> store the transformation fct once!?
        # TODO transformation fct as an attribute -> assign here (e.g. barycentric if complete)

        # TODO automatic make complete?
        # self.origin_poly = self.origin_poly.make_complete()
        # raise ValueError('some transformations only work for complete multi index sets!')

        self._transformation_operator: Optional[np.ndarray] = None

    @property
    def multi_index(self) -> MultiIndexSet:
        """The multi index set of the origin polynomial (and also the target polynomial)."""
        return self.origin_poly.multi_index

    @property
    def grid(self) -> Grid:
        """The grid of the origin polynomial."""
        return self.origin_poly.grid

    # TODO register the transformation classes to the available_transforms dictionary
    # TODO integrate function to retrieve the proper transformation (cf. transformation_utils.py)
    def __init_subclass__(cls, **kwargs):
        """Add a concrete implementation to the registry of available transformations"""
        super().__init_subclass__(**kwargs)
        cls.available_transforms[(cls.origin_type, cls.target_type)] = cls

    # TODO: remove argument. store origin poly once and reuse.
    #  otherwise the user could input an incompatible polynomial
    #  avoids ugly duplicate calls:
    #   l2n_transformation = TransformationLagrangeToNewton(lagrange_poly)
    #   newton_poly = l2n_transformation(lagrange_poly)
    # on the other hand avoid constructing multiple transformation objects for the same basis
    # just for transforming multiple different polynomioals! -> rather check the validity of the input polynomial basis
    def __call__(self, origin_poly: Optional[MultivariatePolynomialSingleABC] = None):
        """Transformation of polynomial.

        This function is called when an instance of transformation is called ``T()`` or ``T(x)``. If called without
        any arguments, the transformation is applied on the origin_poly used to construct the transformation. If an
        argument is passed, the transformation is applied on the instance passed.

        :param origin_poly: (optional)  an instance of the polynomial to be transformed.
        :return: an instance of the polynomial in the target basis.
        """
        if origin_poly is None:
            origin_poly = self.origin_poly
        # TODO check the validity of the basis (basis of the input poly during init, grids match)
        # TODO helper fcts equality of bases ("grid") __eq__
        elif type(origin_poly) != self.origin_type:
            raise TypeError(
                f"Input polynomial type <{type(origin_poly)}> differs from expected polynomial type <{self.origin_type}>"
            )
        # TODO unsafe. user could input wrong polynomial (e.g. different multi_index!)
        #   or even worse: with same multi index but different grid! (undetected!)
        #   -> test input match!
        return self._apply_transformation(origin_poly)

    @property
    @abstractmethod
    def origin_type(self):
        """Abstract container that stores the data type of the origin polynomial.

        This is a placeholder of the ABC, which is overwritten by the concrete implementation.
        """
        pass

    @property
    @abstractmethod
    def target_type(self):
        """Abstract container that stores the data type of the target polynomial.

        This is a placeholder of the ABC, which is overwritten by the concrete implementation.
        """
        pass

    @abstractmethod
    def _get_transformation_operator(self):
        """Abstract container for storing the transformation operator.

        This is a placeholder of the ABC, which is overwritten by the concrete implementation.
        """
        pass

    @property
    def transformation_operator(self) -> OperatorABC:
        """The polynomial basis transformation operator.

        :return: instance of the transformation operator

        Notes
        -----
        The transformation operator once constructed can be reused for transforming other polynomial instance of the
        origin_poly type, which have the same basis (and grid) as the origin_poly, to the target_poly type.
        """
        if self._transformation_operator is None:
            self._transformation_operator = self._get_transformation_operator()
        return self._transformation_operator

    @property
    def _target_indices(self) -> MultiIndexSet:
        """The MultiIndexSet of the target_poly.

        :return: the indices the target polynomial will have
        """

        # NOTE: poly.multi_index and poly.grid.multi_index might not be equal!
        # this is required since e.g. transforming a polynomial in Lagrange basis
        #    into Newton basis possibly "activates" all indices of the basis (grid).

        # TODO more specifically: only all "previous" = lexicographically smaller indices will be active
        #    -> same as completing only the active multi indices.
        # ATTENTION: the multi indices of the basis in use must stay equal!
        return self.origin_poly.grid.multi_index

    def _apply_transformation(self, origin_poly):
        """Transforms the polynomial to the target type of the transformation.

        :param origin_poly: instance of the polynomial to be transformed
        :return: an instance of the transformed polynomial in the target basis
        """
        # TODO discuss: is it meaningful to create a new polynomial instance every time?
        #  perhaps allow an optional output polynomial as parameter and then just update the coefficients?
        # NOTE: construct a new polynomial from the input polynomial in order to copy all relevant internal attributes!
        output_poly = self.target_type.from_poly(origin_poly)
        # ATTENTION: assign the correct expected multi indices!
        output_poly.multi_index = self._target_indices
        # NOTE: only then the coefficients can be assigned, since the shapes need to match with the indices!
        # NOTE: this is calling self.transformation_operator.__matmul__(origin_poly.coeffs)
        target_coeffs = self.transformation_operator @ origin_poly.coeffs
        output_poly.coeffs = target_coeffs
        return output_poly



from typing import Dict, Optional, Set, Tuple

import numba
import numpy as np

# numpy dtypes
# for exponents:

# ATTENTION: when Numba JIT compilation is active. Some functions may crash silently
# without raising the expected errors! (and stack traces, e.g. just SIGSEGV)
# -> perhaps deactivate Numba for DEBUGGING!
DEBUG = True
# DEBUG = False

INT_DTYPE = np.int_
FLOAT_DTYPE = np.float_
DEFAULT_DOMAIN = np.array([-1, 1])

# Numba types. Must match the Numpy dtypes
FLOAT = numba.from_dtype(FLOAT_DTYPE)
# F_TYPE = f8
F_1D = FLOAT[:]
F_2D = FLOAT[:, :]
F_3D = FLOAT[:, :, :]
INT = numba.from_dtype(INT_DTYPE)
I_1D = INT[:]
I_2D = INT[:, :]
B_TYPE = numba.b1

DEFAULT_LP_DEG = 2.0

NOT_FOUND = -1  # meaning: exponent vector is not contained

ARRAY = np.ndarray  # type: ignore
# TYPED_LIST = List[ARRAY]
TYPED_LIST = numba.typed.List
INT_TUPLE = Tuple[int, int]
ARRAY_DICT = Dict[Tuple[int, int, int], Optional[ARRAY]]  # type: ignore
TRAFO_DICT = Dict[Tuple[int, int], ARRAY]  # type: ignore
INT_SET = Set[int]
DICT_TRAFO_TYPE = Tuple[TRAFO_DICT, ARRAY]  # type: ignore
FACTORISED_TRAFO_TYPE = Tuple[ARRAY, ARRAY, ARRAY, ARRAY]  # type: ignore



"""
Abstract base class for the various polynomial base classes.

This module contains the abstract base classes, from which all concrete implementations of polynomial classes shall subclass.
This ensures that all polynomials work with the same interface, so futher features can be formulated without referencing the concrete polynomial implementation. See e.g. :PEP:`3119` for further explanations on that topic.
"""
import abc
from copy import deepcopy
from typing import Any, Optional, Union

import numpy as np

from minterpy.global_settings import ARRAY

from ..grid import Grid
from ..multi_index import MultiIndexSet
from ..utils import _expand_dim, find_match_between
from ..verification import check_shape, check_type_n_values, verify_domain

__all__ = ["MultivariatePolynomialABC", "MultivariatePolynomialSingleABC"]


class MultivariatePolynomialABC(abc.ABC):
    """the most general abstract base class for multivariate polynomials.

    Every data type which needs to behave like abstract polynomial(s) should subclass this class and implement all the abstract methods.

    Attributes
    ----------
    coeffs
    nr_active_monomials
    spatial_dimension
    unisolvent_nodes

    """

    @property
    @abc.abstractmethod
    def coeffs(self) -> ARRAY:  # pragma: no cover
        """Abstract container which stores the coefficients of the polynomial.

        This is a placeholder of the ABC, which is overwritten by the concrete implementation.
        """
        pass

    @coeffs.setter
    def coeffs(self, value):
        pass

    @property
    @abc.abstractmethod
    def nr_active_monomials(self):  # pragma: no cover
        """Abstract container for the number of monomials of the polynomial(s).

        Notes
        -----
        This is a placeholder of the ABC, which is overwritten by the concrete implementation.
        """
        pass

    @property
    @abc.abstractmethod
    def spatial_dimension(self):  # pragma: no cover
        """Abstract container for the dimension of space where the polynomial(s) live on.

        Notes
        -----
        This is a placeholder of the ABC, which is overwritten by the concrete implementation.
        """
        pass

    @property
    @abc.abstractmethod
    def unisolvent_nodes(self):  # pragma: no cover
        """Abstract container for unisolvent nodes the polynomial(s) is(are) defined on.

        Notes
        -----
        This is a placeholder of the ABC, which is overwritten by the concrete implementation.
        """
        pass

    @abc.abstractmethod
    def _eval(self, x) -> Any:  # pragma: no cover
        """Abstract evaluation function.

        Notes
        -----
        This is a placeholder of the ABC, which is overwritten by the concrete implementation.
        """
        pass

    # TODO *args, **kwargs ?! or rather "point" or "x"
    def __call__(self, x) -> Any:
        """Evaluation of the polynomial.

        This function is called, if an instance of the polynomial(s) is called: ``P(x)``

        Parameters
        ----------
        arg : np.ndarray
            Batch array where the polynomial(s) is evaluated. The shape needs to be ``(N,d)`` or ``(N,d,n)``, where ``N`` is the number of points, ``d`` is the dimension of the space and ``n`` is the number of polynomials (if present). Actually, ``arg`` denotes list of `d`-tuples containing the components of the point in space.

        Returns
        -------
        np.ndarray
            The output array of the scalar values of the polynomials the output may be a ndarray when multiple sets of coefficients have been stored

        Notes
        -----
        Internally the concrete implementation of the static method ``_eval`` is called.

        See Also
        --------
        _eval : concrete implementation of the evaluation of the polynomial(s)
        """
        # TODO built in rescaling between user_domain and internal_domain
        #   IDEA: use sklearn min max scaler (transform() and inverse_transform())
        return self._eval(x)

    # anything else any polynomial must support
    # TODO mathematical operations? abstract
    # TODO copy operations. abstract


class MultivariatePolynomialSingleABC(MultivariatePolynomialABC):
    """abstract base class for "single instance" multivariate polynomials

    Attributes
    ----------
    multi_index : MultiIndexSet
        The multi-indices of the multivariate polynomial.
    coeffs : np.ndarray

    internal_domain : array_like
        The domain the polynomial is defined on (basically the domain of the unisolvent nodes).
        Either one-dimensional domain (min,max), a stack of domains for each
        domain with shape (spatial_dimension,2).
    user_domain : array_like
        The domain where the polynomial can be evaluated. This will be mapped onto the ``internal_domain``.
        Either one-dimensional domain ``min,max)`` a stack of domains for each
        domain with shape ``(spatial_dimension,2)``.

    Notes
    -----
    the grid with the corresponding indices defines the "basis" or polynomial space a polynomial is part of.
    e.g. also the constraints for a Lagrange polynomial, i.e. on which points they must vanish.
    ATTENTION: the grid might be defined on other indices than multi_index! e.g. useful for defining Lagrange coefficients with "extra constraints"
    but all indices from multi_index must be contained in the grid!
    this corresponds to polynomials with just some of the Lagrange polynomials of the basis being "active"
    """

    # __doc__ += __doc_attrs__

    _coeffs: Optional[ARRAY] = None

    @staticmethod
    @abc.abstractmethod
    def generate_internal_domain(
        internal_domain, spatial_dimension
    ):  # pragma: no cover
        # no docstring here, since it is given in the concrete implementation
        pass

    @staticmethod
    @abc.abstractmethod
    def generate_user_domain(user_domain, spatial_dimension):  # pragma: no cover
        # no docstring here, since it is given in the concrete implementation
        pass

    # TODO static methods should not have a parameter "self"
    @staticmethod
    @abc.abstractmethod
    def _add(self, other):  # pragma: no cover
        # no docstring here, since it is given in the concrete implementation
        pass

    @staticmethod
    @abc.abstractmethod
    def _sub(self, other):  # pragma: no cover
        # no docstring here, since it is given in the concrete implementation
        pass

    @staticmethod
    @abc.abstractmethod
    def _mul(self, other):  # pragma: no cover
        # no docstring here, since it is given in the concrete implementation
        pass

    @staticmethod
    @abc.abstractmethod
    def _div(self, other):  # pragma: no cover
        # no docstring here, since it is given in the concrete implementation
        pass

    @staticmethod
    @abc.abstractmethod
    def _pow(self, pow):  # pragma: no cover
        # no docstring here, since it is given in the concrete implementation
        pass

    @staticmethod
    def _gen_grid_default(multi_index):
        """Return the default :class:`Grid` for a given :class:`MultiIndexSet` instance.

        For the default values of the Grid class, see :class:`minterpy.Grid`.


        :param multi_index: An instance of :class:`MultiIndexSet` for which the default :class:`Grid` shall be build
        :type multi_index: MultiIndexSet
        :return: An instance of :class:`Grid` with the default optional parameters.
        :rtype: Grid
        """
        return Grid(multi_index)

    def __init__(
        self,
        multi_index: Union[MultiIndexSet, ARRAY],
        coeffs: Optional[ARRAY] = None,
        internal_domain: Optional[ARRAY] = None,
        user_domain: Optional[ARRAY] = None,
        grid: Optional[Grid] = None,
    ):

        if multi_index.__class__ is MultiIndexSet:
            self.multi_index = multi_index
        else:
            # TODO should passing multi indices as ndarray be supported?
            check_type_n_values(multi_index)  # expected ARRAY
            check_shape(multi_index, dimensionality=2)
            self.multi_index = MultiIndexSet(multi_index)

        nr_monomials, spatial_dimension = self.multi_index.exponents.shape
        self.coeffs = coeffs  # calls the setter method and checks the input shape

        if internal_domain is not None:
            check_type_n_values(internal_domain)
            check_shape(internal_domain, (spatial_dimension, 2))
        self.internal_domain = self.generate_internal_domain(
            internal_domain, self.multi_index.spatial_dimension
        )

        if user_domain is not None:  # TODO not better "external domain"?!
            check_type_n_values(user_domain)
            check_shape(user_domain, (spatial_dimension, 2))
        self.user_domain = self.generate_user_domain(
            user_domain, self.multi_index.spatial_dimension
        )

        # TODO make multi_index input optional? otherwise use the indices from grid
        # TODO class method from_grid
        if grid is None:
            grid = self._gen_grid_default(self.multi_index)
        if type(grid) is not Grid:
            raise ValueError(f"unexpected type {type(grid)} of the input grid")

        if not grid.multi_index.is_super_index_set_of(self.multi_index):
            raise ValueError(
                "the multi indices of a polynomial must be a subset of the indices of the grid in use"
            )
        self.grid: Grid = grid
        # weather or not the indices are independent from the grid ("basis")
        # TODO this could be enconded by .active_monomials being None
        self.indices_are_separate: bool = self.grid.multi_index is not self.multi_index
        self.active_monomials: Optional[ARRAY] = None  # 1:1 correspondence
        if self.indices_are_separate:
            # store the position of the active Lagrange polynomials with respect to the basis indices:
            self.active_monomials = find_match_between(
                self.multi_index.exponents, self.grid.multi_index.exponents
            )

    @classmethod
    def from_degree(
        cls,
        spatial_dimension: int,
        poly_degree: int,
        lp_degree: int,
        coeffs: Optional[ARRAY] = None,
        internal_domain: ARRAY = None,
        user_domain: ARRAY = None,
    ):
        """Initialise Polynomial from given coefficients and the default construction for given polynomial degree, spatial dimension and :math:`l_p` degree.

        :param spatial_dimension: Dimension of the domain space of the polynomial.
        :type spatial_dimension: int

        :param poly_degree: The degree of the polynomial, i.e. the (integer) supremum of the :math:`l_p` norms of the monomials.
        :type poly_degree: int

        :param lp_degree: The :math:`l_p` degree used to determine the polynomial degree.
        :type lp_degree: int

        :param coeffs: coefficients of the polynomial. These shall be 1D for a single polynomial, where the length of the array is the number of monomials given by the ``multi_index``. For a set of similar polynomials (with the same number of monomials) the array can also be 2D, where the first axis refers to the monomials and the second axis refers to the polynomials.
        :type coeffs: np.ndarray

        :param internal_domain: the internal domain (factory) where the polynomials are defined on, e.g. :math:`[-1,1]^d` where :math:`d` is the dimension of the domain space. If a ``callable`` is passed, it shall get the dimension of the domain space and returns the ``internal_domain`` as an :class:`np.ndarray`.
        :type internal_domain: np.ndarray or callable
        :param user_domain: the domain window (factory), from which the arguments of a polynomial are transformed to the internal domain. If a ``callable`` is passed, it shall get the dimension of the domain space and returns the ``user_domain`` as an :class:`np.ndarray`.
        :type user_domain: np.ndarray or callable

        """
        return cls(
            MultiIndexSet.from_degree(spatial_dimension, poly_degree, lp_degree),
            coeffs,
            internal_domain,
            user_domain,
        )

    @classmethod
    def from_poly(
        cls,
        polynomial: "MultivariatePolynomialSingleABC",
        new_coeffs: Optional[ARRAY] = None,
    ) -> "MultivariatePolynomialSingleABC":
        """constructs a new polynomial instance based on the properties of an input polynomial

        useful for copying polynomials of other types


        :param polynomial: input polynomial instance defining the properties to be reused
        :param new_coeffs: the coefficients the new polynomials should have. using `polynomial.coeffs` if `None`
        :return: new polynomial instance with equal properties

        Notes
        -----
        The coefficients can also be assigned later.
        """
        p = polynomial
        if new_coeffs is None:  # use the same coefficients
            new_coeffs = p.coeffs

        return cls(p.multi_index, new_coeffs, p.internal_domain, p.user_domain, p.grid)

    # Arithmetic operations:

    def __neg__(self):
        """
        Negation of the polynomial(s).

        :return: new polynomial with negated coefficients.
        :rtype: MultivariatePolynomialSingleABC
        """
        return self.__class__(
            self.multi_index, -self._coeffs, self.internal_domain, self.user_domain
        )

    def __pos__(self):
        """
        Plus signing of polynomial(s).

        :return: the same polynomial
        :rtype: MultivariatePolynomialSingleABC
        """
        return self

    def __add__(self, other):
        """Addition of the polynomial(s) with another given polynomial(s).

        This function is called, if two polynomials are added like ``P1 + P2``.

        :param other: the other polynomial, which is added.
        :type other: MultivariatePolynomialSingleABC

        :return: The result of ``self + other``.
        :rtype: MultivariatePolynomialSingleABC

        Notes
        -----
        Internally it calles the static method ``self._add`` from the concrete implementation.

        See Also
        --------
        _add : concrete implementation of ``__add__``
        """
        if self.__class__ != other.__class__:
            raise NotImplementedError(
                f"Addition operation not implemented for "
                f"'{self.__class__}', '{other.__class__}'"
            )

        result = self._add(self, other)
        return result

    def __sub__(self, other):
        """Substraction of the polynomial(s) with another given polynomial(s).

        This function is called, if two polynomials are substracted like ``P1 - P2``.

        :param other: the other polynomial, which is substracted.
        :type other:  MultivariatePolynomialSingleABC

        :return: The result of ``self - other``.
        :rtype: MultivariatePolynomialSingleABC

        Notes
        -----
        Internally it calles the static method ``self._sub`` from the concrete implementation.

        See Also
        --------
        _sub : concrete implementation of ``__sub__``
        """
        if self.__class__ != other.__class__:
            raise NotImplementedError(
                f"Subtraction operation not implemented for "
                f"'{self.__class__}', '{other.__class__}'"
            )

        result = self._sub(self, other)
        return result

    def __mul__(self, other):
        """Multiplication of the polynomial(s) with another given polynomial(s).

        This function is called, if two polynomials are multiplied like ``P1 * P2``.

        :param other: the other polynomial, which is multiplied.
        :type other:  MultivariatePolynomialSingleABC

        :return: The result of ``self * other``.
        :rtype: MultivariatePolynomialSingleABC

        Notes
        -----
        Internally it calles the static method ``self._mul`` from the concrete implementation.

        See Also
        --------
        _mul : concrete implementation of ``__mul__``
        """
        if self.__class__ != other.__class__:
            raise NotImplementedError(
                f"Multiplication operation not implemented for "
                f"'{self.__class__}', '{other.__class__}'"
            )

        result = self._mul(self, other)
        return result

    def __radd__(self, other):
        """Right sided addition of the polynomial(s) with another given polynomial(s).

        This function is called, if two polynomials are added like ``P1 + P2`` from the right side.

        :param other: The other polynomial, where this instance is added on.
        :type other:  MultivariatePolynomialSingleABC

        :return: The result of ``other + self``.
        :rtype: MultivariatePolynomialSingleABC

        Notes
        -----
        Internally it calles the static method ``self._radd`` from the concrete implementation.

        See Also
        --------
        _radd : concrete implementation of ``__radd__``
        """
        if self.__class__ != other.__class__:
            raise NotImplementedError(
                f"Addition operation not implemented for "
                f"'{self.__class__}', '{other.__class__}'"
            )

        result = self._add(other, self)
        return result

    def __rsub__(self, other):
        """Right sided difference of the polynomial(s) with another given polynomial(s).

        This function is called, if two polynomials are substracted like ``P1 - P2`` from the right side.

        :param other: The other polynomial, where this instance is substracted from.
        :type other: MultivariatePolynomialSingleABC

        :return: The result of ``other - self``.
        :rtype: MultivariatePolynomialSingleABC

        Notes
        -----
        Internally it calles the static method ``self._rsub`` from the concrete implementation.

        See Also
        --------
        _rsub : concrete implementation of ``__rsub__``
        """
        if self.__class__ != other.__class__:
            raise NotImplementedError(
                f"Subtraction operation not implemented for "
                f"'{self.__class__}', '{other.__class__}'"
            )

        result = self._add(-other, self)
        return result

    def __rmul__(self, other):
        """Right sided multiplication of the polynomial(s) with another given polynomial(s).

        This function is called, if two polynomials are multiplied like ``P1*P2`` from the right side.

        :param other: The other polynomial, where this instance is multplied on.
        :type other: MultivariatePolynomialSingleABC

        :return: The result of ``other * self``.
        :rtype: MultivariatePolynomialSingleABC

        Notes
        -----
        Internally it calles the static method ``self._rmul`` from the concrete implementation.

        See Also
        --------
        _rmul : concrete implementation of ``__rmul__``
        """
        if self.__class__ != other.__class__:
            raise NotImplementedError(
                f"Multiplication operation not implemented for "
                f"'{self.__class__}', '{other.__class__}'"
            )

        # TODO Call to the _mul method
        # TODO Return the a new class instance with the result
        return

    # copying
    def __copy__(self):
        """Creates of a shallow copy.

        This function is called, if one uses the top-level function ``copy()`` on an instance of this class.

        :return: The copy of the current instance.
        :rtype: MultivariatePolynomialSingleABC

        See Also
        --------
        copy.copy
            copy operator form the python standard library.
        """
        return self.__class__(
            self.multi_index,
            self._coeffs,
            self.internal_domain,
            self.user_domain,
            self.grid,
        )

    def __deepcopy__(self, mem):
        """Creates of a deepcopy.

        This function is called, if one uses the top-level function ``deepcopy()`` on an instance of this class.

        :return: The deepcopy of the current instance.
        :rtype: MultivariatePolynomialSingleABC

        See Also
        --------
        copy.deepcopy
            copy operator form the python standard library.

        """
        return self.__class__(
            deepcopy(self.multi_index),
            deepcopy(self._coeffs),
            deepcopy(self.internal_domain),
            deepcopy(self.user_domain),
            deepcopy(self.grid),
        )

    @property
    def nr_active_monomials(self):
        """Number of active monomials of the polynomial(s).

        For caching and methods based on switching single monomials on and off, it is distigushed between active and passive monomials, where only the active monomials particitpate on exposed functions.

        :return: Number of active monomials.
        :rtype: int

        Notes
        -----
        This is usually equal to the "amount of coefficients". However the coefficients can also be a 2D array (representing a multitude of polynomials with the same base grid).
        """

        return len(self.multi_index)

    @property
    def spatial_dimension(self):
        """Spatial dimension.

        The dimension of space where the polynomial(s) live on.

        :return: Dimension of domain space.
        :rtype: int

        Notes
        -----
        This is propagated from the ``multi_index.spatial_dimension``.
        """
        return self.multi_index.spatial_dimension

    @property
    def coeffs(self) -> Optional[ARRAY]:
        """Array which stores the coefficients of the polynomial.

        With shape (N,) or (N, p) the coefficients of the multivariate polynomial(s), where N is the amount of monomials and p is the amount of polynomials.

        :return: Array of coefficients.
        :rtype: np.ndarray
        :raise ValueError: Raised if the coeffs are not initialised.

        Notes
        -----
        It is allowed to set the coefficients to `None` to represent a not yet initialised polynomial
        """
        if self._coeffs is None:
            raise ValueError(
                "trying to access an uninitialized polynomial (coefficients are `None`)"
            )
        return self._coeffs

    @coeffs.setter
    def coeffs(self, value: Optional[ARRAY]):
        # setters shall not have docstrings. See numpydoc class example.
        if value is None:
            self._coeffs = None
            return
        check_type_n_values(value)
        if value.shape[0] != self.nr_active_monomials:
            raise ValueError(
                f"the amount of given coefficients <{value.shape[0]}> does not match "
                f"with the amount of monomials in the polynomial <{self.nr_active_monomials}>."
            )
        self._coeffs = value

    @property
    def unisolvent_nodes(self):
        """Unisolvent nodes the polynomial(s) is(are) defined on.

        For definitions of unisolvent nodes see the mathematical introduction.

        :return: Array of unisolvent nodes.
        :rtype: np.ndarray

        Notes
        -----
        This is propagated from from ``self.grid.unisolvent_nodes``.
        """
        return self.grid.unisolvent_nodes

    def _new_instance_if_necessary(
        self, new_grid, new_indices: Optional[MultiIndexSet] = None
    ) -> "MultivariatePolynomialSingleABC":
        """Constructs a new instance only if the multi indices have changed.

        :param new_grid: Grid instance the polynomial is defined on.
        :type new_grid: Grid

        :param new_indices: :class:`MultiIndexSet` instance for the polynomial(s), needs to be a subset of the current ``multi_index``. Default is :class:`None`.
        :type new_indices: MultiIndexSet, optional

        :return: Same polynomial instance if ``grid`` and ``multi_index`` stay the same, otherwise new polynomial instance with the new ``grid`` and ``multi_index``.
        :rtype: MultivariatePolynomialSingleABC
        """
        prev_grid = self.grid
        if new_grid is prev_grid:
            return self
        # grid has changed
        if new_indices is None:
            # the active monomials (and coefficients) stay equal
            new_indices = self.multi_index
            new_coeffs = self._coeffs
        else:
            # also the active monomials change
            prev_indices = self.multi_index
            if not prev_indices.is_sub_index_set_of(new_indices):
                raise ValueError(
                    "an index set of a polynomial can only be expanded, "
                    "but the old indices contain multi indices not present in the new indices."
                )

            # convert the coefficients correctly:
            if self._coeffs is None:
                new_coeffs = None
            else:
                new_coeffs = np.zeros(len(new_indices))
                idxs_of_old = find_match_between(
                    prev_indices.exponents, new_indices.exponents
                )
                new_coeffs[idxs_of_old] = self._coeffs

        new_poly_instance = self.__class__(new_indices, new_coeffs, grid=new_grid)
        return new_poly_instance

    def make_complete(self) -> "MultivariatePolynomialSingleABC":
        """returns a possibly new polynomial instance with a complete multi index set.

        :return: completed polynomial, where additional coefficients setted to zero.
        :rtype: MultivariatePolynomialSingleABC

        Notes
        -----
        - the active monomials stay equal. only the grid ("basis") changes
        - in the case of a Lagrange polynomial this could be done by evaluating the polynomial on the complete grid
        """
        grid_completed = self.grid.make_complete()
        return self._new_instance_if_necessary(grid_completed)

    def add_points(self, exponents: ARRAY) -> "MultivariatePolynomialSingleABC":
        """Extend ``grid`` and ``multi_index``

        Adds points ``grid`` and exponents to ``multi_index`` related to a given set of additional exponents.

        :param exponents: Array of exponents added.
        :type exponents: np.ndarray

        :return: New polynomial with the added exponents.
        :rtype: MultivariatePolynomialSingleABC

        """
        # replace the grid with an independent copy with the new multi indices
        # ATTENTION: the grid might be defined on other indices than multi_index!
        #   but all indices from multi_index must be contained in the grid!
        # -> make sure to add all new additional indices also to the grid!
        grid_new = self.grid.add_points(exponents)
        multi_indices_new = None
        if self.indices_are_separate:
            multi_indices_new = self.multi_index.add_exponents(exponents)
        return self._new_instance_if_necessary(grid_new, multi_indices_new)

    # def make_derivable(self) -> "MultivariatePolynomialSingleABC":
    #     """ convert the polynomial into a new polynomial instance with a "derivable" multi index set
    #  NOTE: not meaningful since derivation requires complete index sets anyway?
    #     """
    #     new_indices = self.multi_index.make_derivable()
    #     return self._new_instance_if_necessary(new_indices)

    def expand_dim(
        self,
        dim: int,
        extra_internal_domain: ARRAY = None,
        extra_user_domain: ARRAY = None,
    ):
        """Expand the spatial dimention.

        Expands the dimension of the domain space of the polynomial by adding zeros to the multi_indices
        (which is equivalent to the multiplication of ones to each monomial).
        Furthermore, the grid is now embedded in the higher dimensional space by pinning the grid arrays to the origin in the additional spatial dimension.

        :param dim: Number of additional dimensions.
        :type dim: int
        """
        expand_dim = dim - self.multi_index.spatial_dimension

        self.multi_index.expand_dim(
            dim
        )  # breaks if dim<spacial_dimension, i.e. expand_dim<0

        grid = self.grid
        new_gen_pts = _expand_dim(grid.generating_points, dim)
        new_gen_vals = _expand_dim(grid.generating_values.reshape(-1, 1), dim)

        self.grid = Grid(self.multi_index, new_gen_pts, new_gen_vals)

        extra_internal_domain = verify_domain(extra_internal_domain, expand_dim)
        self.internal_domain = np.concatenate(
            (self.internal_domain, extra_internal_domain)
        )
        extra_user_domain = verify_domain(extra_user_domain, expand_dim)
        self.user_domain = np.concatenate((self.user_domain, extra_user_domain))

    def partial_diff(self, dim: int, order: int = 1) -> "MultivariatePolynomialSingleABC":
        """Compute the polynomial that is the partial derivative along a dimension of specified order.

        Parameters
        ----------
        dim: spatial dimension along which to take the derivative
        order: order of partial derivative

        Returns
        -------
        a new polynomial instance that represents the partial derivative
        """

        # Guard rails for dim
        if not np.issubdtype(type(dim), np.integer):
            raise TypeError(f"dim <{dim}> must be an integer")

        if dim < 0 or dim >= self.spatial_dimension:
            raise ValueError(f"dim <{dim}> for spatial dimension <{self.spatial_dimension}>"
                             f" should be between 0 and {self.spatial_dimension-1}")

        # Guard rails for order
        if not np.issubdtype(type(dim), np.integer):
            raise TypeError(f"order <{order}> must be a non-negative integer")

        if order < 0:
            raise ValueError(f"order <{order}> must be a non-negative integer")

        return self._partial_diff(self, dim, order)

    def diff(self, order: np.ndarray) -> "MultivariatePolynomialSingleABC":
        """Compute the polynomial that is the partial derivative of a particular order along each dimension.

        Parameters
        ----------
        order: integer array specifying the order of derivative along each dimension

        Returns
        -------
        a new polynomial instance that represents the partial derivative
        """

        # convert 'order' to numpy 1d array if it isn't already. This allows type checking below.
        order = np.ravel(order)

        # Guard rails for order
        if not np.issubdtype(order.dtype.type, np.integer):
            raise TypeError(f"order of derivative <{order}> can only be non-negative integers")

        if np.any(order < 0):
            raise ValueError(f"order of derivative <{order}> cannot have negative values")

        if len(order) != self.spatial_dimension:
            raise ValueError(f"inconsistent number of elements in 'order' <{len(order)}>,"
                             f"expected <{self.spatial_dimension}> corresponding to each spatial dimension")

        return self._diff(self, order)



"""
Utility functions for computing matrices for transformation between canonical, lagrange, and newton basis.

"""
from typing import no_type_check

import numpy as np

from minterpy.core.ABC import OperatorABC, TransformationABC
from minterpy.dds import dds
from minterpy.global_settings import ARRAY, DEBUG, FLOAT_DTYPE
from minterpy.jit_compiled_utils import compute_vandermonde_n2c
from minterpy.schemes.barycentric.precomp import (
    _build_lagrange_to_newton_bary,
    _build_newton_to_lagrange_bary,
)
from minterpy.schemes.matrix_operator import MatrixOperator
from minterpy.utils import eval_newton_monomials

# NOTE: avoid looping over a numpy array! e.g. for j in np.arange(num_monomials):
# see: # https://stackoverflow.com/questions/10698858/built-in-range-or-numpy-arange-which-is-more-efficient


# todo simplify names


def invert_triangular(triangular_matrix: np.ndarray) -> np.ndarray:
    # FIXME: triangular inversion is not working! required when using barycentric transforms?
    # i, j = triangular_matrix.shape  # square matrix
    # inverted_matrix = solve_triangular(triangular_matrix, np.identity(i))
    inverted_matrix = np.linalg.inv(triangular_matrix)
    return inverted_matrix


def _build_n2l_array(grid, multi_index=None, require_invertible: bool = False) -> ARRAY:
    # NOTE: the indices might be different from the ones used in the grid!
    # -> just some "active" Lagrange polynomials
    if multi_index is None or require_invertible:
        # NOTE: the transformation matrix needs to be square for the inversion
        # even if some of the multi indices are "inactive",
        # the unisolvent nodes (grid) need to correspond to the multi indices!
        multi_index = grid.multi_index

    exponents = multi_index.exponents
    unisolvent_nodes = grid.unisolvent_nodes
    generating_points = grid.generating_points
    # NOTE: the shape of unisolvent_nodes and exponents might be different! -> non square transformation matrix
    transformation_matrix = eval_newton_monomials(
        unisolvent_nodes,
        exponents,
        generating_points,
        verify_input=DEBUG,
        triangular=True,
    )
    return transformation_matrix


def _build_newton_to_lagrange_naive(
    transformation: TransformationABC,
) -> MatrixOperator:
    """computes the Newton to Lagrange transformation given by an array

    SPECIAL PROPERTY: the evaluation of any polynomial on unisolvent nodes yields
        the Lagrange coefficients of this polynomial
        (for the Lagrange basis defined implicitly by these unisolvent nodes)
    -> evaluating every Newton polynomial (<-> the monomials forming the Newton basis) on all interpolation points,
        naturally yields the operator transforming the Newton coefficients into Lagrange coefficients

     NOTE: it is inefficient to compute this by inversion:
         newton_to_lagrange = inv(lagrange_to_newton) # based on DDS

    special property: half of the values will always be 0 (lower triangular matrix).

    :param require_invertible: weather or not the output matrix should be square
    """
    grid = transformation.grid
    transformation_matrix = _build_n2l_array(
        grid, transformation.origin_poly.multi_index
    )
    transformation_operator = MatrixOperator(transformation, transformation_matrix)
    return transformation_operator


def build_l2n_matrix_dds(grid):
    num_monomials = len(grid.multi_index)
    lagr_coeff_matrix = np.eye(num_monomials, dtype=FLOAT_DTYPE)
    tree = grid.tree
    lagrange_to_newton = dds(lagr_coeff_matrix, tree)
    return lagrange_to_newton


def _build_lagrange_to_newton_naive(
    transformation: TransformationABC,
) -> MatrixOperator:
    """computes the Lagrange to Newton transformation given by an array

    NOTE: each column of the L2N transformation matrix
        corresponds to the Newton coefficients of the respective Lagrange polynomial.

    NOTE: computing the L2N matrix could be done with the DDS scheme, but only if the multi indices are complete.
        in this case however it is more efficient to use the barycentric transformations right away.
    for the case that the indices are not complete the matrix must be computed by inverting the N2L transformation.
    """
    newton_to_lagrange = _build_n2l_array(transformation.grid, require_invertible=True)
    transformation_matrix = invert_triangular(newton_to_lagrange)
    transformation_operator = MatrixOperator(transformation, transformation_matrix)
    return transformation_operator


def _build_c2n_array(transformation) -> ARRAY:
    multi_index = transformation.grid.multi_index
    num_monomials = len(multi_index)
    V_n2c = np.ones((num_monomials, num_monomials), dtype=FLOAT_DTYPE)
    compute_vandermonde_n2c(
        V_n2c, transformation.grid.unisolvent_nodes, multi_index.exponents
    )  # computes the result "in place"
    tree = transformation.grid.tree
    c2n = dds(V_n2c, tree)
    return c2n


def _build_n2c_array(transformation: TransformationABC) -> ARRAY:
    # TODO achieve by calling inverse(n2c_operator)
    return invert_triangular(_build_c2n_array(transformation))


# TODO own module?


@no_type_check
def _build_newton_to_lagrange_operator(
    transformation: TransformationABC,
) -> OperatorABC:
    """constructs the Newton to Lagrange transformation operator

    use the barycentric transformation if the indices are complete!
    TODO find solution for the case that the multi indices are separate from the grid indices

    :param transformation: the Transformation instance
        with the fixed polynomial defining the unisolvent nodes to perform the transformation on
    :return: the transformation operator from Newton to Lagrange basis
    """
    grid = transformation.grid
    complete_indices = grid.multi_index.is_complete
    identical_indices = not transformation.origin_poly.indices_are_separate
    if complete_indices and identical_indices:
        # use barycentric transformation
        transformation_operator = _build_newton_to_lagrange_bary(transformation)
    else:  # use "naive" matrix transformation format
        transformation_operator = _build_newton_to_lagrange_naive(transformation)

    return transformation_operator


@no_type_check
def _build_lagrange_to_newton_operator(
    transformation: TransformationABC,
) -> OperatorABC:
    """constructs the Lagrange to Newton transformation operator

    use the barycentric transformation if the indices are complete!
    TODO find solution for the case that the multi indices are separate from the grid indices

    NOTE: it is inefficient to compute this by inversion:
        newton_to_lagrange = inv(lagrange_to_newton) # based on DDS

    :param transformation: the Transformation instance
        with the fixed polynomial defining the unisolvent nodes to perform the transformation on
    :return: the transformation operator from Newton to Lagrange basis
    """
    grid = transformation.grid
    complete_indices = grid.multi_index.is_complete
    identical_indices = not transformation.origin_poly.indices_are_separate
    if complete_indices and identical_indices:
        # use barycentric transformation
        transformation_operator = _build_lagrange_to_newton_bary(transformation)
    else:  # use "naive" matrix transformation format
        transformation_operator = _build_lagrange_to_newton_naive(transformation)
    return transformation_operator


# TODO test these transformations:
def _build_canonical_to_newton_operator(
    transformation: TransformationABC,
) -> MatrixOperator:
    return MatrixOperator(transformation, _build_c2n_array(transformation))


def _build_newton_to_canonical_operator(
    transformation: TransformationABC,
) -> MatrixOperator:
    return MatrixOperator(transformation, _build_n2c_array(transformation))


def _build_lagrange_to_canonical_operator(
    transformation: TransformationABC,
) -> OperatorABC:
    lagrange_to_newton = _build_lagrange_to_newton_operator(transformation)
    newton_to_canonical = _build_newton_to_canonical_operator(transformation)
    return newton_to_canonical @ lagrange_to_newton


def _build_canonical_to_lagrange_operator(
    transformation: TransformationABC,
) -> OperatorABC:
    newton_to_lagrange = _build_newton_to_lagrange_operator(transformation)
    canonical_to_newton = _build_canonical_to_newton_operator(transformation)
    return newton_to_lagrange @ canonical_to_newton



"""
Module of the MultiIndexSet class
"""
from copy import deepcopy
from typing import Optional

import numpy as np

from minterpy.global_settings import ARRAY, INT_DTYPE
from minterpy.jit_compiled_utils import (
    all_indices_are_contained,
    have_lexicographical_ordering,
)

from .utils import (
    _expand_dim,
    _get_poly_degree,
    get_exponent_matrix,
    insert_lexicographically,
    is_lexicographically_complete,
    make_complete,
    verify_lp_deg,
)
from .verification import check_shape, check_values

__all__ = ["MultiIndexSet"]


# TODO implement (set) comparison operations based on the multi index utils (>=, == ...)
class MultiIndexSet:
    """Class representation of the set of multi indices for a polynomial.

    The instances of this class provide storrage for the exponents of a multi variate
    polynomial independently of the base assumed in the polynomial space. Only
    the polynomial degree (w.r.t. a given `l_p` norm) as well as the dimension of the space are used to build the set of exponents.

    Attributes
    ----------
    exponents
    lp_degree
    exponents_completed
    is_complete
    spatial_dimension
    """

    def __init__(self, exponents: ARRAY, lp_degree=None):

        # Check and assign the exponents
        exponents = np.require(exponents, dtype=INT_DTYPE)
        check_shape(exponents, dimensionality=2)
        if not have_lexicographical_ordering(exponents):
            raise ValueError(
                "The multi-index set must be lexicographically ordered from "
                "the last to the first column."
            )
        self._exponents: ARRAY = exponents

        # TODO compute properly, max norm of the exponents?
        # while _get_poly_degree(exponents, __lp_degree) > poly_degree: ...
        self._lp_degree = verify_lp_deg(lp_degree)

        # Compute the polynomial degree given the exponents and lp-degree
        self.poly_degree = _get_poly_degree(exponents, self._lp_degree)

        self._is_complete: Optional[bool] = None
        # for avoiding to complete the exponents multiple times
        self._exponents_completed: Optional[ARRAY] = None

    @classmethod
    def from_degree(cls, spatial_dimension: int, poly_degree: int, lp_degree=None):
        if type(poly_degree) is not int:
            raise TypeError("only integer polynomial degrees are supported.")
        if type(spatial_dimension) is not int:
            raise TypeError("spatial dimension must be given as integer.")
        lp_degree = verify_lp_deg(lp_degree)
        exponents = get_exponent_matrix(spatial_dimension, poly_degree, lp_degree)
        return cls(exponents, lp_degree=lp_degree)

    def expand_dim(self, dim):
        # TODO avoid transpose
        # print("self.__exponents",self.__exponents)
        # print("self.__exponents.T",self.__exponents.T)
        # print("_expand_dim(self.__exponents.T,dim)",_expand_dim(self.__exponents.T,dim))
        self._exponents = _expand_dim(self._exponents, dim)

    @property
    def exponents(
        self,
    ):
        """Array of exponents.

        :return: A 2D :class:`ndarray` which stores the exponents, where the first axis refers to the explicit exponent and the second axis refers to the variable which is powerd by the entry at this array position. These exponents are always lexicographically ordered w.r.t. the first axis.

        :rtype: np.ndarray
        """
        # read only: must not be altered since transformation matrices etc. depends on this

        return self._exponents

    @property
    def exponents_completed(self):
        """
        The completed version of the exponent array.

        :return: A complete array of the ``exponents``, i.e. in lexicographical ordering, every exponent which is missing in ``exponents``, will be filled with the right entry.
        :rtype: np.ndarray

        """
        # NOTE: exponents which are not complete ("with holes") cause problems with DDS, evaluation...
        # compute and store a completed version of the indices for these operations!
        if self._exponents_completed is None:  # lazy evaluation
            if self.is_complete:
                self._exponents_completed = self._exponents
            else:
                self._exponents_completed = make_complete(
                    self._exponents, self.lp_degree
                )
        return self._exponents_completed

    @property
    def is_complete(self):
        """Returns :class:`True` if the ``exponent`` array is complete.

        :rtype: bool
        """
        # NOTE: exponents which are not complete ("with holes") cause problems with DDS, evaluation...
        if self._is_complete is None:  # lazy evaluation
            self._is_complete = is_lexicographically_complete(self.exponents)
            if self._is_complete:
                self._exponents_completed = self._exponents
        return self._is_complete

    @property
    def lp_degree(self):
        """The degree for the :math:`l_p`-norm used to define the polynomial degree.

        :return: The lp-degree of the :math:`l_p`-norm, which is used to define the polynomial degree.
        :rtype: int

        :raise ValueError: If a given lp_degree has ``lp_degree<=0``.
        """
        return self._lp_degree

    @lp_degree.setter
    def lp_degree(self, lp_degree):
        # TODO is a setter really meaningful for this attribute?
        #  should rather be computed from the exponents and be fixed afterwards
        if lp_degree <= 0.0:
            raise ValueError(
                f"The lp_degree needs to be a positive value! <{lp_degree}> given."
            )
        self._lp_degree = lp_degree

    @property
    def spatial_dimension(self):
        """The dimentsion of the domain space.

         :return: The dimension of the space, where a polynomial described by this ``multi_index`` lives on. It is equal to the number of powers in each exponent vector.

        :rtype: int

        """
        return self._exponents.shape[1]

    def __str__(self):
        return "\n".join(["MultiIndexSet", str(self._exponents)])

    def __repr__(self):
        return self.__str__()

    def __len__(self):
        # returns the number of multi_indices
        return self._exponents.shape[0]

    def union(self):
        """This function is not implemented yet.

        :raise NotImplementedError: If this function is called.

        """
        raise NotImplementedError("MultiIndexSet.union() is not implemented yet.")

    def __add__(self):
        """This function is not implemented yet.

        :raise NotImplementedError: If this function is called.

        """
        raise NotImplementedError("MultiIndexSet.__add__() is not implemented yet.")

    def ordering(self, order):
        """This function is not implemented yet.

        :raise NotImplementedError: If this function is called.

        """
        raise NotImplementedError("MultiIndexSet.ordering() is not implemented yet.")

    # def complete(self, order):
    #     raise NotImplementedError("MultiIndexSet.complete() is not implemented yet.")

    def copy_private_attributes(self, new_instance):
        """Copys the private attributes to another instance of :class:`MultiIndexSet`

        .. todo::
            - use the ``__copy__`` hook instead!

        """
        new_instance._is_complete = self._is_complete
        new_instance._exponents_completed = self._exponents_completed

    # copying
    def __copy__(self):
        """Creates of a shallow copy.

        This function is called, if one uses the top-level function ``copy()`` on an instance of this class.

        :return: The copy of the current instance.
        :rtype: MultiIndexSet

        See Also
        --------
        copy.copy
            copy operator form the python standard library.
        """
        # TODO also copy tree if available
        # TODO maybe use always deepcopy since exponents will be always nested?!
        new_instance = self.__class__(
            self._exponents, self._lp_degree
        )
        self.copy_private_attributes(new_instance)

        return new_instance

    def __deepcopy__(self, memo):
        """Creates of a deepcopy.

        This function is called, if one uses the top-level function ``deepcopy()`` on an instance of this class.

        :return: The deepcopy of the current instance.
        :rtype: MultiIndexSet

        See Also
        --------
        copy.deepcopy
            copy operator form the python standard library.
        """
        new_instance = self.__class__(
            deepcopy(self.exponents),
            deepcopy(self._lp_degree),
        )
        self.copy_private_attributes(new_instance)

        return new_instance

    def contains_these_exponents(self, vectors: ARRAY) -> bool:
        """Checks if this instance contains a given set of exponents.

        :param vectors: Exponents to be checked.
        :type vectors: np.ndarray

        """
        return all_indices_are_contained(vectors, self._exponents)

    def is_sub_index_set_of(self, super_set: "MultiIndexSet") -> bool:
        """Checks if this instance is a subset of the given instance of :class:`MultiIndexSet`.

        :param super_set: Superset to be checked.
        :type super_set: MultiIndexSet

        .. todo::
            - use comparison hooks for this functionality
        """
        return super_set.contains_these_exponents(self.exponents)

    def is_super_index_set_of(self, sub_set: "MultiIndexSet") -> bool:
        """Checks if this instance is a super of the given instance of :class:`MultiIndexSet`.

        :param sub_set: Subset to be checked.
        :type super_set: MultiIndexSet

        .. todo::
            - use comparison hooks for this functionality
        """
        return self.contains_these_exponents(sub_set.exponents)

    def _new_instance_if_necessary(self, new_exponents: ARRAY) -> "MultiIndexSet":
        """constructs a new instance only if the exponents are different

        .. todo::
            - this should be a separate function rather than a member function.
            - this should use the comparison hooks instead of attribute checking.
        """
        old_exponents = self._exponents
        if new_exponents is old_exponents:
            return self
        new_instance = self.__class__(new_exponents, self.lp_degree)
        # TODO add to the completed exponents and re-complete again
        if self._exponents_completed is not None:
            _exponents_completed = insert_lexicographically(
                self._exponents_completed, new_exponents
            )
            if (
                _exponents_completed is not self._exponents_completed
            ):  # some exponents have been added:
                # make complete again!
                _exponents_completed = make_complete(
                    _exponents_completed, self.lp_degree
                )
            new_instance._exponents_completed = _exponents_completed
            # also set the exponents of the new_instance
            # TODO avoid redundancy. why store both _exponents and _exponents_completed?
            # new_instance._exponents = _exponents_completed
        return new_instance

    def add_exponents(self, exponents: ARRAY) -> "MultiIndexSet":
        exponents = np.require(exponents, dtype=INT_DTYPE)
        check_values(exponents)
        exponents = exponents.reshape(-1, self.spatial_dimension)
        """Insert a set if exponents lexicographically into the exponents of this instance.

        :param exponents: Array of exponents to be inserted.
        :type exponents: np.ndarray

        :return: New instance of :class:`MultiIndexSet` (if necessary), where the additional exponents are inserted uniquely.
        :rtype: MultiIndexSet


        .. todo::
            - shall be renamed to ``insert_exponents`` or ``insert_multi_index``
        """
        # convert input to 2D in expected shape
        #  NOTE: the insertion is expected to return the same identical array instance
        #  if all exponents are already contained!
        # -> no need to check for inclusion first!
        new_exponents = insert_lexicographically(self._exponents, exponents)
        return self._new_instance_if_necessary(new_exponents)

    def make_complete(self) -> "MultiIndexSet":
        """Completion of this instance.

        Bulid a new instance of :class:`MultiIndexSet`, where the exponents are completed, i.e. in lexicographical ordering, there will be no exponent missing.

        :return: Completed version of this instance
        :rtype: MultiIndexSet

        .. todo::
            - since this returns a copy of this instance, maybe this shall be given as a separate function.
            - why not an implace version of that?

        """
        # ATTENTION: the make_complete() fct should not be called with already complete idx sets (inefficient!)
        # should return the same object if already complete (avoid creation of identical instance)
        if self.is_complete:
            return self
        new_exponents = self.exponents_completed  # compute if necessary!
        new_instance = self.__class__(new_exponents)  # re-compute degree etc.
        # NOTE: avoid checking for completeness again!
        new_instance._is_complete = True
        new_instance._exponents_completed = new_exponents
        return new_instance

    # TODO make_derivable(): add (only) partial derivative exponent vectors,
    #  NOTE: not meaningful since derivation requires complete index sets anyway?



"""
Concrete implementation of Identity transformation when there is no basis change.

Notes
-----
This class is needed for completeness reasons and also to enable the high-level helper functions for polynomial basis
transformations to handle transformations between identical basis.
"""

from copy import copy

import numpy as np

from minterpy.core.ABC import MultivariatePolynomialABC, TransformationABC
from minterpy.schemes.matrix_operator import MatrixOperator

__all__ = ["Identity"]


def _build_identity_transformation_operator(transformation):
    return MatrixOperator(transformation, np.identity(len(transformation.multi_index)))


class Identity(TransformationABC):
    """Transformation between same basis."""

    origin_type = MultivariatePolynomialABC
    target_type = MultivariatePolynomialABC
    _get_transformation_operator = _build_identity_transformation_operator

    def _apply_transformation(self, origin_poly):
        return copy(origin_poly)  # no need to compute dot product



"""
testing module for canonical_polynomial.py

The subclassing is not tested here, see tesing module `test_polynomial.py`
"""
import numpy as np
import pytest
from conftest import (
    SEED,
    LpDegree,
    MultiIndices,
    NrPoints,
    NrPolynomials,
    NrSimilarPolynomials,
    PolyDegree,
    SpatialDimension,
    BatchSizes,
    assert_polynomial_almost_equal,
    build_rnd_coeffs,
    build_rnd_points,
    build_random_newton_polynom,
)
from numpy.testing import assert_, assert_almost_equal
from minterpy.global_settings import INT_DTYPE
from minterpy.utils import eval_newton_polynomials
from minterpy import Grid

from minterpy import NewtonPolynomial, NewtonToCanonical, CanonicalToNewton


def test_eval(MultiIndices, NrPoints, NrPolynomials):
    """Test the evaluation of Newton polynomials."""

    coeffs = build_rnd_coeffs(MultiIndices, NrPolynomials)
    poly = NewtonPolynomial(MultiIndices, coeffs)
    pts = build_rnd_points(NrPoints, MultiIndices.spatial_dimension)

    # Evaluate
    res = poly(pts)

    trafo_n2c = NewtonToCanonical(poly)
    canon_poly = trafo_n2c()
    groundtruth = canon_poly(pts)
    assert_almost_equal(res, groundtruth)


def test_eval_batch(MultiIndices, NrPolynomials, BatchSizes):
    """Test the evaluation on Newton polynomials in batches of query points."""

    #TODO: This is a temporary test as the 'batch_size' parameter is not
    #      opened in the higher-level interface, i.e., 'newton_poly(xx)'

    # Create a random coefficient values
    newton_coeffs = build_rnd_coeffs(MultiIndices, NrPolynomials)
    grid = Grid(MultiIndices)
    generating_points = grid.generating_points
    exponents = MultiIndices.exponents

    # Create test query points
    xx = build_rnd_points(421, MultiIndices.spatial_dimension)

    # Evaluate the polynomial in batches
    yy_newton = eval_newton_polynomials(
        xx, newton_coeffs, exponents, generating_points, batch_size=BatchSizes
    )

    # Create a reference results from canonical polynomial evaluation
    newton_poly = NewtonPolynomial(MultiIndices, newton_coeffs)
    canonical_poly = NewtonToCanonical(newton_poly)()
    yy_canonical = canonical_poly(xx)

    # Assert
    assert_almost_equal(yy_newton, yy_canonical)


def test_partial_diff(SpatialDimension, PolyDegree, LpDegree):
    newton_poly = build_random_newton_polynom(SpatialDimension, PolyDegree, LpDegree)

    # Check partial derivative on each dimension by comparing it with Canonical partial derivatives
    for dim in range(SpatialDimension):
        trafo_n2c = NewtonToCanonical(newton_poly)
        canon_poly = trafo_n2c()
        can_diff_poly = canon_poly.partial_diff(dim)
        trafo_c2n = CanonicalToNewton(can_diff_poly)
        newt_can_diff_poly = trafo_c2n()

        newt_diff_poly = newton_poly.partial_diff(dim)

        assert_polynomial_almost_equal(newt_can_diff_poly, newt_diff_poly)

def test_diff(SpatialDimension, PolyDegree, LpDegree):
    newton_poly = build_random_newton_polynom(SpatialDimension, PolyDegree, LpDegree)

    # A derivative of order zero along all dimensions should be equivalent to the same polynomial
    zero_order_diff_newt = newton_poly.diff(np.zeros(SpatialDimension, dtype=INT_DTYPE))
    assert_polynomial_almost_equal(zero_order_diff_newt, newton_poly)

    # Comparing the gradient with that computed in Canonical basis
    trafo_n2c = NewtonToCanonical(newton_poly)
    canon_poly = trafo_n2c()
    diff_order = np.ones(SpatialDimension, dtype=INT_DTYPE)
    can_diff_poly = canon_poly.diff(diff_order)
    trafo_c2n = CanonicalToNewton(can_diff_poly)
    newt_can_diff_poly = trafo_c2n()

    newt_diff_poly = newton_poly.diff(diff_order)

    assert_polynomial_almost_equal(newt_can_diff_poly, newt_diff_poly)



"""
testing module for the polynomial submodule.

Here only the execution/initialization is tested, not the correct values. For value testing, see the testing module of the concrete implementations.
"""
import numpy as np
import pytest
from conftest import assert_call
from numpy.testing import assert_

from minterpy import (
    CanonicalPolynomial,
    LagrangePolynomial,
    MultiIndexSet,
    NewtonPolynomial,
)
from minterpy.core.ABC import MultivariatePolynomialSingleABC

classes = [CanonicalPolynomial, NewtonPolynomial, LagrangePolynomial]
class_ids = tuple(cls.__name__ for cls in classes)


@pytest.fixture(params=classes, ids=class_ids)
def Polynom(request):
    return request.param


# test init


def test_init_polynomials(Polynom):
    assert_(issubclass(Polynom, MultivariatePolynomialSingleABC))

    mi = MultiIndexSet.from_degree(2, 1, 1)
    coeffs = np.arange(len(mi), dtype=float)
    assert_call(Polynom, mi, coeffs)



"""
Module of the NewtonPolynomial class

.. todo::
    - implement staticmethods for Newton polynomials (or at least transform them to another base).
"""
import numpy as np

from minterpy.global_settings import DEBUG
from minterpy.utils import eval_newton_polynomials
from minterpy.dds import dds
from .utils import deriv_newt_eval

from ..core.ABC.multivariate_polynomial_abstract import MultivariatePolynomialSingleABC
from ..core.verification import verify_domain

__all__ = ["NewtonPolynomial"]


def dummy():
    """Placeholder function.

    .. warning::
      This feature is not implemented yet!
    """
    # move dummy to util?
    raise NotImplementedError("This feature is not implemented yet.")


def newton_eval(newton_poly, x):
    """Evaluation function in Newton base.

    This is a wrapper for the evaluation function in Newton base.

    :param newton_poly: The :class:`NewtonPolynomial` which is evaluated.
    :type newton_poly: NewtonPolynomial
    :param x: The argument(s) the Newton polynomial shall be evaluated. The input shape needs to be ``(N,dim)``, where ``N`` refered to the number of points and ``dim`` refers to the dimension of the domain space, i.e. the coordinates of the argument vector.


    .. todo::
        - check right order of axes for ``x``.

    See Also
    --------
    newt_eval : comcrete implementation of the evaluation in Newton base.

    """
    return eval_newton_polynomials(
        x,
        newton_poly.coeffs,
        newton_poly.multi_index.exponents,
        newton_poly.grid.generating_points,
        verify_input=DEBUG,
    )


def _newton_partial_diff(poly: "NewtonPolynomial", dim: int, order: int) -> "NewtonPolynomial":
    """ Partial differentiation in Newton basis.

    Notes
    -----
    Performs a transformation from Lagrange to Newton using DDS.
    """
    spatial_dim = poly.multi_index.spatial_dimension
    deriv_order_along = [0]*spatial_dim
    deriv_order_along[dim] = order
    return _newton_diff(poly, deriv_order_along)

def _newton_diff(poly: "NewtonPolynomial", order: np.ndarray) -> "NewtonPolynomial":
    """ Partial differentiation in Newton basis.

    Notes
    -----
    Performs a transformation from Lagrange to Newton using DDS.
    """

    # When you evaluate the derivatives on the unisolvent nodes, you get the coefficients for
    # the derivative polynomial in Lagrange basis.
    lag_coeffs = deriv_newt_eval(poly.grid.unisolvent_nodes, poly.coeffs,
                                 poly.grid.multi_index.exponents,
                                 poly.grid.generating_points, order)

    # DDS returns a 2D array, converting it to 1d
    newt_coeffs = dds(lag_coeffs, poly.grid.tree).reshape(-1)

    return NewtonPolynomial(coeffs=newt_coeffs, multi_index=poly.multi_index,
                              grid=poly.grid)

# TODO redundant
newton_generate_internal_domain = verify_domain
newton_generate_user_domain = verify_domain


class NewtonPolynomial(MultivariatePolynomialSingleABC):
    """Datatype to describe polynomials in Newton base.

    For a definition of the Newton base, see the mathematical introduction.

    .. todo::
        - provide a short definition of this base here.

    Attributes
    ----------
    coeffs
    nr_active_monomials
    spatial_dimension
    unisolvent_nodes


    """

    # Virtual Functions
    _add = staticmethod(dummy)
    _sub = staticmethod(dummy)
    _mul = staticmethod(dummy)
    _div = staticmethod(dummy)
    _pow = staticmethod(dummy)
    _eval = newton_eval

    _partial_diff = staticmethod(_newton_partial_diff)
    _diff = staticmethod(_newton_diff)

    generate_internal_domain = staticmethod(newton_generate_internal_domain)
    generate_user_domain = staticmethod(newton_generate_user_domain)



"""
Here we store the core utilities of `minterpy`.
"""
from __future__ import annotations

from typing import Iterable, no_type_check

import numpy as np
from math import ceil

from minterpy.global_settings import DEFAULT_LP_DEG, INT_DTYPE
from minterpy.jit_compiled_utils import (
    fill_match_positions,
    index_is_contained,
    lex_smaller_or_equal,
)
from minterpy.utils import cartesian_product, lp_norm, lp_sum

# if TYPE_CHECKING:
#    from .tree import MultiIndexTree


def _get_poly_degree(exponents: np.ndarray, lp_degree: float) -> int:
    """Get the polynomial degree from a multi-index set for a given lp-degree.
    
    Parameters
    ----------
    exponents : np.ndarray
        An array of exponents of a multi-index set.
    
    lp_degree : float
        The lp-degree of the multi-index set.

    Returns
    -------
    int
        The polynomial degree from the multi-index set
        for the given lp-degree.
    """
    norms = lp_norm(exponents, lp_degree, axis=1)
    # NOTE: math.ceil() returns int, np.ceil() returns float
    return ceil(np.max(norms))

def get_exponent_matrix(
    spatial_dimension: int, poly_degree: int, lp_degree: float | int
) -> np.ndarray:
    """
    Generate exponent matrix.

    :param spatial_dimension: Dimension of the domain space.
    :type spatial_dimension: int

    :param poly_degree: Polynomial degree described by the resulting exponent matrix.
    :type poly_degree: int

    :param lp_degree: Degree of the used lp-norm. This can be any integer bigger than one or `np.inf`.
    :type lp_degree: int or float

    :return: List of exponent arrays for all monomials, whose lp-norm of the exponents is smaller or equal to the given `poly_degree`. The list is given as a `np.ndarray` with the shape `(number_of_monomials, spatial_dimension)` and the list is lexicographically ordered.
    :rtype: np.ndarray

    """
    # Validate poly_degree, allow float if finite with integral value
    if isinstance(poly_degree, float):
        if not poly_degree.is_integer():
            raise ValueError(
                f"poly_degree needs to be a whole number! <{poly_degree} given."
            )

    if poly_degree == 0:
        right_choices = np.zeros((1, spatial_dimension), dtype=INT_DTYPE)

        return right_choices

    if lp_degree == np.inf:
        right_choices = cartesian_product(
            *[np.arange(poly_degree + 1, dtype=INT_DTYPE)] * spatial_dimension
        )
    else:
        candidates_without_diag = cartesian_product(
            *[np.arange(poly_degree, dtype=INT_DTYPE)] * spatial_dimension
        )
        candidates = np.vstack(
            (
                candidates_without_diag,
                np.diag([INT_DTYPE(poly_degree)] * spatial_dimension),
            )
        )
        cond = lp_sum(candidates, lp_degree) <= poly_degree ** lp_degree
        right_choices = candidates[cond]
    lex_idx = np.lexsort(right_choices.T)
    return right_choices[lex_idx]


NORM_FCT = lp_norm


def _gen_multi_index_exponents_recur(m, n, gamma, gamma2, lp_degree):
    """DEPRECATED. only for reference. TODO remove
    NOTE: this is slow for larger problem instances!

    build multi index "gamma" depending on lp_degree
    NOTE: multi indices are being called "alpha" in the newest interpolation paper

    BUG: lp = infinity does not create complete equidistant grids ((n+1)^m entries)
    very memory inefficient implementation!

    :param m:
    :param n:
    :param gamma:
    :param gamma2:
    :param lp_degree:
    :return:
    """

    # NOTE: these (and only these) copy operations are required! (tested)
    gamma2 = gamma2.copy()
    gamma0 = gamma.copy()
    gamma0[0, m - 1] += 1

    out = []
    norm = NORM_FCT(gamma0.reshape(-1), lp_degree)
    if norm < n and m > 1:
        o1 = _gen_multi_index_exponents_recur(m - 1, n, gamma, gamma, lp_degree)
        o2 = _gen_multi_index_exponents_recur(m, n, gamma0, gamma0, lp_degree)
        out = np.concatenate([o1, o2], axis=0)
    elif norm < n and m == 1:
        out = np.concatenate(
            [gamma2, _gen_multi_index_exponents_recur(m, n, gamma0, gamma0, lp_degree)],
            axis=0,
        )
    elif norm == n and m > 1:
        out = np.concatenate(
            [
                _gen_multi_index_exponents_recur(m - 1, n, gamma, gamma, lp_degree),
                gamma0,
            ],
            axis=0,
        )
    elif norm == n and m == 1:
        out = np.concatenate([gamma2, gamma0], axis=0)
    elif norm > n:
        norm_ = NORM_FCT(gamma.reshape(-1), lp_degree)
        if norm_ < n and m > 1:
            for j in range(1, m):
                gamma0 = gamma  # TODO simplify
                gamma0[j - 1] = gamma0[j - 1] + 1  # gamm0 -> 1121 broken
                if NORM_FCT(gamma0.reshape(-1), lp_degree) <= n:
                    gamma2 = np.concatenate(
                        [
                            gamma2,
                            _gen_multi_index_exponents_recur(
                                j, n, gamma0, gamma0, lp_degree
                            ),
                        ],
                        axis=0,
                    )
            out = gamma2
        elif m == 1:
            out = gamma2
        elif norm_ <= n:
            out = gamma
    return out


def _gen_multi_index_exponents(spatial_dimension, poly_degree, lp_degree):
    """
    creates the array of exponents for the MultiIndex class
    DEPRECATED: reference implementations for tests TODO remove
    """
    gamma_placeholder = np.zeros((1, spatial_dimension), dtype=INT_DTYPE)
    exponents = _gen_multi_index_exponents_recur(
        spatial_dimension, poly_degree, gamma_placeholder, gamma_placeholder, lp_degree
    )
    return exponents


def _expand_dim(grid_nodes, target_dim, point_pinned=None):
    """
    Expansion of a given array to a given dimension, where the additional dimensions will filled with tiles of given elements

    Parameters
    ----------
    grid_nodes : array_like (shape=(len_arr,dim))
        Array which shall be expanded.
    target_dim : np.int
        Dimension up to the array shall be expanded to. Needs to be bigger or equal than the current dimension of array.
    point_pinned : optional np.ndarray
    """
    grid_len, dim = grid_nodes.shape
    grid_dtype = grid_nodes.dtype
    if target_dim < dim:
        # TODO maybe build a reduce function which removes dims where all exps are 0
        raise ValueError(f"Can't expand grid from dim {dim} to dim {target_dim}.")
    if target_dim == dim:
        return grid_nodes
    num_expand_dim = target_dim - dim
    if point_pinned is None:
        new_dim_exps = np.zeros((grid_len, num_expand_dim), dtype=grid_dtype)
    else:
        point_pinned = np.atleast_1d(point_pinned)
        if len(point_pinned) is not num_expand_dim:
            raise ValueError(
                f"Given point_pinned {point_pinned} has not enough elements to fill the extra dimensions! <{num_expand_dim}> required."
            )
        new_dim_exps = np.require(
            np.tile(point_pinned, (grid_len, 1)), dtype=grid_dtype
        )
    return np.concatenate((grid_nodes, new_dim_exps), axis=1)


def iterate_indices(indices: np.ndarray | Iterable[np.ndarray]) -> Iterable[np.ndarray]:
    if isinstance(indices, np.ndarray) and indices.ndim == 1:  # just a single index
        yield indices
    else:  # already iterable as is:
        yield from indices


def gen_partial_derivatives(exponent_vector: np.ndarray):
    """yields the exponent vectors of all partial derivatives of a given exponent vector

    NOTE: all exponent vectors "smaller by 1"
    """
    spatial_dimension = len(exponent_vector)
    for m in range(spatial_dimension):
        exponent_in_dim = exponent_vector[m]
        if exponent_in_dim == 0:  # partial derivative is 0
            continue
        partial_derivative = exponent_vector.copy()
        partial_derivative[m] = exponent_in_dim - 1
        yield partial_derivative


def gen_missing_derivatives(indices: np.ndarray) -> Iterable[np.ndarray]:
    """yields all the partial derivative exponent vectors missing from the given set of multi indices

    ATTENTION: the input indices must have lexicographical ordering
    ATTENTION: duplicate partial derivatives will be generated
    """
    nr_exponents, spatial_dimension = indices.shape
    for n in reversed(
        range(nr_exponents)
    ):  # start with the lexicographically biggest index
        exp_vect = indices[n, :]
        for deriv_vect in gen_partial_derivatives(
            exp_vect
        ):  # all vectors "smaller by 1"
            # NOTE: looking for the index starting from the last index would be faster TODO
            indices2search = indices[:n, :]
            if not index_is_contained(indices2search, deriv_vect):
                yield deriv_vect


def is_lexicographically_complete(indices: np.ndarray) -> bool:
    """tells weather an array of indices contains all possible lexicographically smaller indices

    ATTENTION: the input indices must have lexicographical ordering
    :returns False if there is a missing multi index vector "smaller by 1"
    """
    for _ in gen_missing_derivatives(indices):
        return False
    return True


@no_type_check
def list_insert_single(
    list_of_indices: list[np.ndarray],
    index2insert: np.ndarray,
    check_for_inclusion: bool = True,
):
    """inserts a single index into a given list of indices maintaining lexicographical ordering"""
    # TODO check length of single index to insert!
    if list_of_indices is None:
        raise TypeError
    nr_of_indices = len(list_of_indices)
    if nr_of_indices == 0:
        list_of_indices.insert(0, index2insert)
        return

    insertion_idx = (
        nr_of_indices  # default: insert at the last position, ATTENTION: -1 not working
    )
    for i, contained_index in enumerate(list_of_indices):
        if lex_smaller_or_equal(index2insert, contained_index):
            insertion_idx = i
            break

    # no insertion when an equal entry exists already
    # TODO check_for_inclusion
    if not np.array_equal(contained_index, index2insert):
        list_of_indices.insert(insertion_idx, index2insert)


@no_type_check
def to_index_list(indices: np.ndarray | Iterable[np.ndarray]) -> list[np.ndarray]:
    if type(indices) is list:
        return indices  # already is a list
    list_of_indices = list(iterate_indices(indices))  # include all existing indices
    return list_of_indices


def to_index_array(list_of_indices: list[np.ndarray]) -> np.ndarray:
    # NOTE: shape is: (N, m)
    index_array = np.array(list_of_indices, dtype=INT_DTYPE)
    return index_array


# def list_based(func):
#     """ TODO decorator for working with list of indices rather than with static numpy arrays"""
#     @functools.wraps(func)
#     def list_based_wrapper(*args, **kwargs):
#         to_index_list(
#         value = func(*args, **kwargs)
#         to_index_array(
#         return value
#     return list_based_wrapper


def insert_lexicographically(
    indices: list[np.ndarray] | np.ndarray,
    indices2insert: Iterable[np.ndarray] | None,
) -> np.ndarray:
    """inserts possibly multiple index vectors into a given array of indices maintaining lexicographical ordering

    exploits ordering for increased efficiency:
     inserts one vector after another to maintain ordering

    NOTE: is expected to return the same identical array instance
     if all exponents are already contained!
    """
    if indices2insert is None:
        return indices
    nr_exponents = len(indices)
    list_of_indices = None  # avoid creating a list when there is no index to insert
    for i, index2insert in enumerate(iterate_indices(indices2insert)):
        if i == 0:  # initialise list
            list_of_indices = to_index_list(indices)
        list_insert_single(list_of_indices, index2insert)
    if (
        list_of_indices is None or len(list_of_indices) == nr_exponents
    ):  # len(list_of_indices) or len(indices)
        # no index has been inserted.
        # ATTENTION: return the previous array in order to easily compare for equality!
        return indices
    index_array = to_index_array(list_of_indices)
    return index_array


def insert_partial_derivatives(list_of_indices, exponent_vector):
    for deriv_vect in gen_partial_derivatives(
        exponent_vector
    ):  # all vectors "smaller by 1"
        list_insert_single(list_of_indices, deriv_vect)


def make_derivable(indices: np.ndarray) -> np.ndarray:
    """inserts all missing multi index vectors "smaller by one" """
    list_of_indices: list[int] = []
    nr_exponents, spatial_dimension = indices.shape
    for i in reversed(range(nr_exponents)):  # start with the biggest multi index
        contained_exponent_vector = indices[i, :]
        list_insert_single(list_of_indices, contained_exponent_vector)
        insert_partial_derivatives(list_of_indices, contained_exponent_vector)
    index_array = to_index_array(list_of_indices)
    return index_array


def make_complete(indices: np.ndarray, lp_degree: float = None) -> np.ndarray:
    """Make a given array of exponents complete.

    :param indices: The exponent array to be completed.
    :type indices: np.ndarray
    :param lp_degree: lp-degree for the completation. Optional, the default is given by `minterpy.DEFAULT_LP_DEG`.
    :type lp_degree: int (optional)

    :return: Completed version of the input exponents.
    :rtype: np.ndarray

    """
    if lp_degree is None:
        lp_degree = DEFAULT_LP_DEG
    poly_degree = _get_poly_degree(indices, lp_degree)
    spatial_dimension = indices.shape[-1]
    return get_exponent_matrix(spatial_dimension, poly_degree, lp_degree)


def find_match_between(
    smaller_idx_set: np.ndarray, larger_idx_set: np.ndarray
) -> np.ndarray:
    """find the positions of each multi_index of the smaller set in the larger set

    NOTE: both sets are required to be ordered lexicographically!
    NOTE: the smaller set is required to be a subset of the larger set
    NOTE: this function is required for 'rebasing' a polynomial onto complete multi indices (map the coefficients)
    """
    nr_exp_smaller, spatial_dimension = smaller_idx_set.shape
    positions = np.empty(nr_exp_smaller, dtype=INT_DTYPE)
    fill_match_positions(larger_idx_set, smaller_idx_set, positions)
    return positions


def verify_lp_deg(lp_degree):
    if lp_degree is None:
        lp_degree = DEFAULT_LP_DEG
    # TODO remove
    # elif lp_degree != DEFAULT_LP_DEG:
    #     if lp_degree % 1.0 != 0.0:
    #         raise ValueError('the given l_p-degree cannot be interpreted as integer')
    #     lp_degree = int(lp_degree)
    return lp_degree



"""
Module defining the base class for MultiIndexTree.
"""


from __future__ import annotations

from typing import TYPE_CHECKING

from minterpy.dds import (
    compile_problem_sizes,
    compile_splits,
    compile_subtree_sizes,
    precompute_masks,
)
from minterpy.global_settings import ARRAY, ARRAY_DICT  # noqa

if TYPE_CHECKING:
    # https://stackoverflow.com/questions/39740632/python-type-hinting-without-cyclic-imports
    from .grid import Grid
    from .multi_index import MultiIndexSet


__all__ = ["MultiIndexTree"]


class MultiIndexTree:
    """Base class for MultiIndexTree

    Attributes
    ----------
    multi_index : MultiIndex

    grid : Grid

    split_positions : List

    subtree_sizes : List

    problem_sizes : List

    """

    #  TODO prevent dynamic attribute assignment (-> safe memory)
    # __slots__ = ["multi_index", "split_positions", "subtree_sizes", "stored_masks", "generating_points"]

    def __init__(self, grid: Grid):
        multi_index = grid.multi_index
        if not multi_index.is_complete:
            raise ValueError(
                "trying to use the divided difference scheme (multi index tree) "
                "with incomplete multi indices, "
                "but DDS only works for complete multi indices (without 'holes')."
            )

        self.grid = grid

        exponents = multi_index.exponents
        nr_exponents, spatial_dimension = exponents.shape
        # NOTE: the tree structure ("splitting") depends on the exponents
        # in each dimension of the sorted multi index array
        # pre-compute and store where the splits appear in the exponent array
        # this implicitly defines the "nodes" of the tree
        # TODO compute on demand? NOTE: tree is being constructed only on demand (DDS)
        # TODO reverse the dim order of all
        #  (NOTE: then the "dim_idx" will then be counter intuitive: 0 for highest dimension...)
        self.split_positions = compile_splits(exponents)
        # also store the size of all nodes = how many exponent entries belong to this split
        # in combination with the positions of all appearing splits
        # the sizes fully determine the structure of the multi index tree
        # (position and amount of children etc.)
        self.subtree_sizes = compile_subtree_sizes(nr_exponents, self.split_positions)

        self.problem_sizes = compile_problem_sizes(self.subtree_sizes)

        # TODO improvement: also "pre-compute" more of the recursion through the tree,
        #  avoid computing the node indices each time
        self._stored_masks: ARRAY_DICT | None = None

    @property
    def multi_index(self) -> MultiIndexSet:
        """Returns the multi index set of the grid used to construct the tree.

        :return: the multi index set of the grid
        """
        return self.grid.multi_index

    @property
    def stored_masks(self) -> ARRAY_DICT:
        """Returns the stored masks of the tree.

        :return: correspondencies between the left and right nodes of the tree
        """
        # the intermediary results required for DDS
        # TODO remove when regular DDS functionality is no longer required (together with the dds module)
        if self._stored_masks is None:  # lazy evaluation
            # based on the splittings one can compute all required correspondences
            # between nodes in the left and the right of the tree
            # this mapping can then be used to compute the nD DDS efficiently
            exponents = self.multi_index.exponents
            self._stored_masks = precompute_masks(
                self.split_positions, self.subtree_sizes, exponents
            )
        return self._stored_masks



"""
Module containing several numba optimized functions.
"""

import numpy as np
from numba import b1, njit, void

from minterpy.global_settings import (
    B_TYPE,
    F_1D,
    F_2D,
    FLOAT,
    I_1D,
    I_2D,
    INT,
    FLOAT_DTYPE,
    NOT_FOUND,
)


@njit(void(F_2D, F_2D, I_2D, F_2D), cache=True)
def can_eval_mult(x_multiple, coeffs, exponents, result_placeholder):
    """Naive evaluation of polynomials in canonical basis.

    - ``m`` spatial dimension
    - ``k`` number of points
    - ``N`` number of monomials
    - ``p`` number of polynomials

    :param x_multiple: numpy array with coordinates of points where polynomial is to be evaluated.
                       The shape has to be ``(k x m)``.
    :param coeffs: numpy array of polynomial coefficients in canonical basis. The shape has to be ``(N x p)``.
    :param exponents: numpy array with exponents for the polynomial. The shape has to be ``(N x m)``.
    :param result_placeholder: placeholder numpy array where the results of evaluation are stored.
                               The shape has to be ``(k x p)``.

    Notes
    -----
    This is a naive evaluation; a more numerically accurate approach would be to transform to Newton basis and
    using the newton evaluation scheme.

    Multiple polynomials in the canonical basis can be evaluated at once by having a 2D coeffs array. It is assumed
    that they all have the same set of exponents.

    """
    nr_coeffs, nr_polys = coeffs.shape
    r = result_placeholder
    nr_points, _ = x_multiple.shape
    for i in range(nr_coeffs):  # each monomial
        exp = exponents[i, :]
        for j in range(nr_points):  # evaluated on each point
            x = x_multiple[j, :]
            monomial_value = np.prod(np.power(x, exp))
            for k in range(nr_polys):  # reuse intermediary results
                c = coeffs[i, k]
                r[j, k] += c * monomial_value


# NOTE: the most "fine grained" functions must be defined first
# in order for Numba to properly infer the function types

@njit(void(F_1D, I_2D, F_2D, I_1D, F_2D, F_1D), cache=True)  # O(Nm)
def eval_newton_monomials_single(
    x_single,
    exponents,
    generating_points,
    max_exponents,
    products_placeholder,
    monomials_placeholder,
) -> None:
    """Precomputes the value of all given Newton basis polynomials at a point.

    Core of the fast polynomial evaluation algorithm.
    - ``m`` spatial dimension
    - ``N`` number of monomials
    - ``n`` maximum exponent in each dimension

    :param x_single: coordinates of the point. The shape has to be ``m``.
    :param exponents: numpy array with exponents for the polynomial. The shape has to be ``(N x m)``.
    :param generating_points: generating points used to generate the grid. The shape is ``(n x m)``.
    :param max_exponents: array with maximum exponent in each dimension. The shape has to be ``m``.
    :param products_placeholder: a numpy array for storing the (chained) products.
    :param monomials_placeholder: a numpy array of length N for storing the values of all Newton basis polynomials.

    Notes
    -----
    - This is a Numba-accelerated function.
    - The function precompute all the (chained) products required during Newton evaluation for a single query point
      with complexity of ``O(mN)``.
    - The (pre-)computation of Newton monomials is coefficient agnostic.
    - Results are stored in the placeholder arrays. The function returns None.
    """

    # NOTE: the maximal exponent might be different in every dimension,
    #    in this case the matrix becomes sparse (towards the end)
    # NOTE: avoid index shifting during evaluation (has larger complexity than pre-computation!)
    #    by just adding one empty row in front. ATTENTION: these values must not be accessed!
    #    -> the exponents of each monomial ("alpha") then match the indices of the required products

    # Create the products matrix
    m = exponents.shape[1]
    for i in range(m):
        max_exp_in_dim = max_exponents[i]
        x_i = x_single[i]
        prod = 1.0
        for j in range(max_exp_in_dim):  # O(n)
            # TODO there are n+1 1D grid values, the last one will never be used!?
            p_ij = generating_points[j, i]
            prod *= x_i - p_ij
            # NOTE: shift index by one
            exponent = j + 1  # NOTE: otherwise the result type is float
            products_placeholder[exponent, i] = prod

    # evaluate all Newton polynomials. O(Nm)
    N = exponents.shape[0]
    for j in range(N):
        # the exponents of each monomial ("alpha")
        # are the indices of the products which need to be multiplied
        newt_mon_val = 1.0  # required as multiplicative identity
        for i in range(m):
            exp = exponents[j, i]
            # NOTE: an exponent of 0 should not cause a multiplication
            # (inefficient, numerical instabilities)
            if exp > 0:
                newt_mon_val *= products_placeholder[exp, i]
        monomials_placeholder[j] = newt_mon_val
    #NOTE: results have been stored in the numpy arrays. no need to return anything.


@njit(void(F_2D, I_2D, F_2D, I_1D, F_2D, F_2D, B_TYPE), cache=True)
def eval_newton_monomials_multiple(
    xx: np.ndarray,
    exponents: np.ndarray,
    generating_points: np.ndarray,
    max_exponents: np.ndarray,
    products_placeholder: np.ndarray,
    monomials_placeholder: np.ndarray,
    triangular: bool
) -> None:
    """Evaluate the Newton monomials at multiple query points.

    The following notations are used below:

    - :math:`m`: the spatial dimension of the polynomial
    - :math:`p`: the (maximum) degree of the polynomial in any dimension
    - :math:`n`: the number of elements in the multi-index set (i.e., monomials)
    - :math:`\mathrm{nr_{points}}`: the number of query (evaluation) points
    - :math:`\mathrm{nr_polynomials}`: the number of polynomials with different
      coefficient sets of the same multi-index set

    :param xx: numpy array with coordinates of points where polynomial is to be evaluated.
              The shape has to be ``(k x m)``.
    :param exponents: numpy array with exponents for the polynomial. The shape has to be ``(N x m)``.
    :param generating_points: generating points used to generate the grid. The shape is ``(n x m)``.
    :param max_exponents: array with maximum exponent in each dimension. The shape has to be ``m``.
    :param products_placeholder: a numpy array for storing the (chained) products.
    :param monomials_placeholder: placeholder numpy array where the results of evaluation are stored.
                               The shape has to be ``(k x p)``.
    :param triangular: whether the output will be of lower triangular form or not.
                       -> will skip the evaluation of some values
    :return: the value of each Newton polynomial at each point. The shape will be ``(k x N)``.

    Notes
    -----
    - This is a Numba-accelerated function.
    - The memory footprint for evaluating the Newton monomials iteratively
       with a single query point at a time is smaller than evaluating all
       the Newton monomials on all query points.
       However, when multiplied with multiple coefficient sets,
       this approach will be faster.
    - Results are stored in the placeholder arrays. The function returns None.
    """

    n_points = xx.shape[0]

    # By default, all exponents are "active" unless xx are unisolvent nodes
    active_exponents = exponents
    # Iterate each query points and evaluate the Newton monomials
    for idx in range(n_points):

        x_single = xx[idx, :]

        # Get the row view of the monomials placeholder;
        # this would be the evaluation results of a single query point
        monomials_placeholder_single = monomials_placeholder[idx]

        if triangular:
            # TODO: Refactor this, this is triangular because the monomials
            #       are evaluated at the unisolvent nodes, otherwise it won't
            #       be and the results would be misleading.
            # When evaluated on unisolvent nodes, some values will be a priori 0
            n_active_polys = idx + 1
            # Only some exponents are active
            active_exponents = exponents[:n_active_polys, :]
            # IMPORTANT: initialised empty. set all others to 0!
            monomials_placeholder_single[n_active_polys:] = 0.0
            # Only modify the non-zero entries
            monomials_placeholder_single = \
                monomials_placeholder_single[:n_active_polys]

        # Evaluate the Newton monomials on a single query point
        # NOTE: Due to "view" access,
        # the whole 'monomials_placeholder' will be modified
        eval_newton_monomials_single(
            x_single,
            active_exponents,
            generating_points,
            max_exponents,
            products_placeholder,
            monomials_placeholder_single
        )


@njit(void(F_2D, F_2D, I_2D), cache=True)
def compute_vandermonde_n2c(V_n2c, nodes, exponents):
    """Computes the Vandermonde matrix.

    - ``m`` spatial dimension
    - ``N`` number of monomials

    :param V_n2c: the placeholder array to store the Vandermonde matrix. The shape has to be ``(N x N)``.
    :param nodes: the unisolvent nodes
    :param exponents:  numpy array with exponents for the polynomial. The shape has to be ``(N x m)``.

    """
    num_monomials, spatial_dimension = exponents.shape
    for i in range(num_monomials):
        for j in range(1, num_monomials):
            for d in range(spatial_dimension):
                V_n2c[i, j] *= nodes[i, d] ** exponents[j, d]


@njit(b1(I_1D, I_1D), cache=True)
def lex_smaller_or_equal(index1: np.ndarray, index2: np.ndarray) -> bool:
    """Compares whether multi-index 1 is lexicographically smaller than or equal to multi-index 2.

    - ``m`` spatial dimension

    :param index1: a multi-index entry of shape ``m``.
    :param index2: another multi-index entry.
    :return: ``True`` if ``index1 <= index2`` (lexicographically), otherwise ``False``.

    """
    spatial_dimension = len(index1)
    for m in range(spatial_dimension - 1, -1, -1):  # from last to first dimension
        if index1[m] > index2[m]:
            return False
        if index1[m] < index2[m]:
            return True
    return True  # all equal


@njit(B_TYPE(I_2D), cache=True)
def have_lexicographical_ordering(indices: np.ndarray) -> bool:
    """Checks if an array of indices is ordered lexicographically

    - ``m`` spatial dimension
    - ``N`` number of monomials

    :param indices: array of multi-indices with shape ``(N x m)``.
    :return: ``True`` if indices are lexicographically ordered, ``False`` otherwise.

    """
    nr_exponents, spatial_dimension = indices.shape
    if nr_exponents <= 1:
        return True
    i1 = indices[0, :]
    for n in range(1, nr_exponents):
        i2 = indices[n, :]
        if lex_smaller_or_equal(i2, i1):
            return False
        if np.all(i1 == i2):  # duplicates are not allowed
            return False
        i1 = i2
    return True


@njit(INT(I_2D, I_1D), cache=True)
def get_match_idx(indices: np.ndarray, index: np.ndarray) -> int:
    """Finds the position of a multi index entry within an exponent matrix.

    - ``m`` spatial dimension
    - ``N`` number of monomials

    :param indices: array of multi indices with shape ``(N x m)``.
    :param index: one multi index entry with shape ``m``.
    :return: if ``index`` is present in ``indices``, the position (array index) where it is found is returned,
             otherwise a global constant ``NOT_FOUND`` is returned.

    Notes
    -----
    Exploits the lexicographical order of the indices to abort early -> not testing all indices.
    Time complexity: O(mN).

    """
    nr_exponents, spatial_dimension = indices.shape
    if nr_exponents == 0:
        return NOT_FOUND
    m = len(index)
    if m != spatial_dimension:
        raise ValueError("dimensions do not match.")
    out = NOT_FOUND
    for i in range(nr_exponents):  # O(N)
        contained_index = indices[i, :]
        if lex_smaller_or_equal(index, contained_index):  # O(m)
            # i is now pointing to the (smallest) index which is lexicographically smaller or equal
            # the two indices are equal iff the contained index is also smaller or equal than the query index
            # NOTE: testing for equality directly is faster, but only in the case of equality (<- rare!)
            #   most of the times the index won't be smaller and the check can be performed with fewer comparisons
            is_equal = lex_smaller_or_equal(contained_index, index)
            if is_equal:  # found the position of the index
                out = i
            break  # stop looking (an even bigger index cannot be equal)
    return out


@njit(B_TYPE(I_2D, I_1D), cache=True)
def index_is_contained(indices: np.ndarray, index: np.ndarray) -> bool:
    """Checks if a multi index entry is present in a set of indices.

    - ``m`` spatial dimension
    - ``N`` number of monomials

    :param indices: array of multi indices with shape ``(N x m)``.
    :param index: one multi index entry with shape ``m``.
    :return: ``True`` if the ``index`` is present in ``indices``, ``False`` otherwise.

    Notes
    -----
    Exploits the lexicographical order of the indices to abort early -> not testing all indices.

    """
    return get_match_idx(indices, index) != NOT_FOUND


@njit(B_TYPE(I_2D, I_2D), cache=True)
def all_indices_are_contained(subset_indices: np.ndarray, indices: np.ndarray) -> bool:
    """Checks if a set of indices is a subset of (or equal to) another set of indices.

    :param subset_indices: one set of multi indices.
    :param indices: another set of multi indices.
    :return: ``True`` if ``subset_indices`` is a subset or equal to ``indices``, ``False`` otherwise.

    Notes
    -----
    Exploits the lexicographical order of the indices to abort early -> not testing all indices.
    """
    nr_exp, dim = indices.shape
    nr_exp_subset, dim_subset = subset_indices.shape
    if nr_exp == 0 or nr_exp_subset == 0:
        raise ValueError("empty index set")
    if dim != dim_subset:
        raise ValueError("dimensions do not match.")
    if nr_exp < nr_exp_subset:
        return False

    # return True when all candidate indices are contained
    match_idx = -1
    for i in range(nr_exp_subset):
        candidate_index = subset_indices[i, :]
        indices2search = indices[match_idx + 1 :, :]  # start from the next one
        match_idx = get_match_idx(indices2search, candidate_index)
        if match_idx == NOT_FOUND:
            return False
    return True


@njit(void(I_2D, I_2D, I_1D), cache=True)
def fill_match_positions(larger_idx_set, smaller_idx_set, positions):
    """Finds matching positions (array indices) for multi index entries in two multi indices.

    :param larger_idx_set: the larger set of multi indices
    :param smaller_idx_set: the smaller set of multi indices
    :param positions: an array with positions (array indices) of multi index entries in `larger_idx_set`` that matches
                      each of the entry in ``smaller_idx_set``.

    """
    search_pos = -1
    nr_exp_smaller, spatial_dimension = smaller_idx_set.shape
    for i in range(nr_exp_smaller):
        idx1 = smaller_idx_set[i, :]
        while 1:
            search_pos += 1
            idx2 = larger_idx_set[search_pos, :]
            if lex_smaller_or_equal(idx1, idx2) and lex_smaller_or_equal(idx2, idx1):
                # NOTE: testing for equality directly is faster, but only in the case of equality (<- rare!)
                #   most of the times the index won't be smaller and the check can be performed with fewer comparisons
                positions[i] = search_pos
                break
